import copy
import os
import numpy as np
from astropy.io import fits
from sunpy import map as smap
import warnings
import sys

warnings.simplefilter("ignore")

stokesval = {'1': 'I', '2': 'Q', '3': 'U', '4': 'V', '-1': 'RR', '-2': 'LL', '-3': 'RL', '-4': 'LR', '-5': 'XX',
             '-6': 'YY', '-7': 'XY', '-8': 'YX'}



def headerfix(header, PC_coor=True):
    '''
	this code fixes the header problem of fits out from CASA 5.4+ which leads to a streched solar image.
    Setting PC_coor equal to True will reset the rotation matrix.
    '''

    keys2remove = []
    for k in header:
        if k.upper().startswith('PC'):
            if not k.upper().startswith('PC0'):
                pcidxs = k.upper().replace('PC', '')
                hd_ = 'PC0' + pcidxs
                keys2remove.append(k)
                if PC_coor:
                    pcidx0, pcidx1 = pcidxs.split('_')
                    if pcidx0 == pcidx1:
                        header[hd_] = 1.0
                    else:
                        header[hd_] = 0.0
                else:
                    header[hd_] = header[k]
    for k in keys2remove:
        header.remove(k)
    return header


def headerparse(header):
    '''
        get axis index of polarization
    '''

    ndim = header['NAXIS']
    stokesinfo = {'axis': None, 'headernew': {}}
    keys2analy = ['NAXIS', 'CTYPE', 'CRVAL', 'CDELT', 'CRPIX', 'CUNIT']
    for dim in range(1, ndim + 1):
        k = 'CTYPE{}'.format(dim)
        if header[k].startswith('STOKES'):
            stokesinfo['axis'] = dim
    if stokesinfo['axis'] is not None:
        dim = stokesinfo['axis']
        for k in keys2analy:
            k_ = '{}{}'.format(k, dim)
            stokesinfo['headernew']['{}{}'.format(k, 4)] = header[k_]
    return stokesinfo


def headersqueeze(header, data):
    '''
        Only 1D, 2D, or 3D images are currently supported by
        the astropy fits compression.
        This code remove single-dimensional entries from the n-dimensional
        image data array and update fits header
    '''
    dshape = data.shape
    ndim = data.ndim
    ## nsdim: numbers of single-dimensional entries
    nsdim = np.count_nonzero(np.array(dshape) == 1)
    ## nonsdim: non-single-dimensional entries
    nonsdim = ndim - nsdim
    if nsdim == 0:
        return header, data
    else:
        keys2chng = ['NAXIS', 'CTYPE', 'CRVAL', 'CDELT', 'CRPIX', 'CUNIT']  # ,'PC01_', 'PC02_', 'PC03_', 'PC04_']
        idx_nonsdim = 0
        for idx, dim in enumerate(dshape[::-1]):
            # if dim>1: continue
            if dim > 1:
                idx_nonsdim = idx_nonsdim + 1
            for k in keys2chng:
                k_ = '{}{}'.format(k, idx + 1)
                v = header[k_]
                header.remove(k_)
                if dim > 1:
                    k_new = '{}{}'.format(k, idx_nonsdim)
                    header[k_new] = v
                else:
                    if k == 'CTYPE' and v.startswith('STOKES'):
                        header['STOKES'] = header['CRVAL{}'.format(idx + 1)]

        idx_nonsdim1 = 0

        for idx1, dim1 in enumerate(dshape[::-1]):
            if dim1 > 1:
                idx_nonsdim1 = idx_nonsdim1 + 1
            idx_nonsdim2 = 0
            for idx2, dim2 in enumerate(dshape[::-1]):
                if dim2 > 1:
                    idx_nonsdim2 = idx_nonsdim2 + 1
                k_ = 'PC{:02d}_{:d}'.format(idx1 + 1, idx2 + 1)
                if k_ in header.keys():
                    v = header[k_]
                    header.remove(k_)
                    if dim1 > 1 and dim2 > 1:
                        k_new = 'PC{:02d}_{:d}'.format(idx_nonsdim1, idx_nonsdim2)
                        header[k_new] = v

        header['NAXIS'] = nonsdim
        data = np.squeeze(data)
        return header, data


def get_bdinfo(freq, bw):
    """
    get band information from center frequencies and band widths.

    Parameters
    ----------
    freq : array_like
        an array of the center frequencies of all frequency bands in Hz
    bw: array_like
        an array of the band widths of all frequency bands in Hz

    Returns
    -------
    fbounds : `dict`
        A dict of band information
    """
    fghz = freq / 1.e9
    bwghz = bw / 1.e9
    bounds_lo = fghz - bwghz / 2.0
    bounds_hi = fghz + bwghz / 2.0
    bounds_all = np.hstack([bounds_lo, bounds_hi[-1]])
    fbounds = {'cfreqs': fghz, 'cfreqs_all': fghz, 'bounds_lo': bounds_lo,
               'bounds_hi': bounds_hi, 'bounds_all': bounds_all}
    return fbounds


def read(filepath, hdus=None, verbose=False, **kwargs):
    """
    Read a fits file.

    Parameters
    ----------
    filepath : `str`
        The fits file to be read.
    hdus: `int` or iterable
        The HDU indexes to read from the file.
    verbose: `bool`
        if verbose

    Returns
    -------
    pairs : `list`
        A list of (data, header) tuples

    Notes
    -----
    This routine reads all the HDU's in a fits file and returns a list of the
    data and a FileHeader instance for each one.

    Also all comments in the original file are concatenated into a single
    "comment" key in the returned FileHeader.
    """
    import collections

    with fits.open(filepath, ignore_blank=True) as hdulist:
        if hdus is not None:
            if isinstance(hdus, int):
                hdulist = hdulist[hdus]
            elif isinstance(hdus, collections.Iterable):
                hdulist = [hdulist[i] for i in hdus]

        hdulist = fits.hdu.HDUList(hdulist)
        for h in hdulist:
            h.verify('silentfix+warn')

        meta = {}
        for i, hdu in enumerate(hdulist):
            try:
                ndim = hdu.data.ndim
                header = hdu.header
                slc = [slice(None)] * ndim
                freq_axis = None
                pol_axis = None
                _ = None
                npol = 1
                nfreq_fits = 1
                for idx in range(ndim):
                    v = header['CTYPE{}'.format(idx + 1)]
                    if v.startswith('FREQ'):
                        freq_axis = ndim - (idx + 1)
                        nfreq_fits = header['NAXIS{}'.format(idx + 1)]
                    if v.startswith('STOKES'):
                        pol_axis = ndim - (idx + 1)
                        npol = header['NAXIS{}'.format(idx + 1)]
                if freq_axis is not None:
                    slc[freq_axis] = slice(0, 1)
                    meta['ref_cfreqs'] = (hdu.header['CRVAL{}'.format(ndim - freq_axis)] + hdu.header[
                        'CDELT{}'.format(ndim - freq_axis)] * np.arange(
                        hdu.header['NAXIS{}'.format(ndim - freq_axis)]))
                    meta['ref_freqdelts'] = np.ones_like(meta['ref_cfreqs']) \
                                            * hdu.header['CDELT{}'.format(ndim - freq_axis)]
                else:
                    meta['ref_cfreqs'] = np.array([hdu.header['RESTFRQ']])

                if pol_axis is not None:
                    slc[pol_axis] = slice(0, 1)
                    meta['pol_idxs'] = (hdu.header['CRVAL{}'.format(ndim - pol_axis)] + hdu.header[
                        'CDELT{}'.format(ndim - pol_axis)] * np.arange(
                        hdu.header['NAXIS{}'.format(ndim - pol_axis)]))
                    meta['pol_names'] = [stokesval['{0:d}'.format(int(p))] for p in meta['pol_idxs']]
                if _ is not None:
                    slc[_] = slice(0, 1)
                hdu.data[np.isnan(hdu.data)] = 0.0
                rmap = smap.Map(np.squeeze(hdu.data[tuple(slc)]), hdu.header)
                data = hdu.data.copy()
                meta['header'] = hdu.header.copy()
                meta['refmap'] = rmap  # this is a sunpy map of the first slice
                meta['naxis'] = ndim
                meta['hgln_axis'] = ndim - 1  # solar X
                meta['nx'] = header['NAXIS1']
                meta['hglt_axis'] = ndim - 2  # solar Y
                meta['ny'] = header['NAXIS2']
                meta['freq_axis'] = freq_axis
                meta['nfreq'] = nfreq_fits
                meta['pol_axis'] = pol_axis
                meta['npol'] = npol
                break
            except Exception as e:
                if verbose:
                    if hasattr(e, 'message'):
                        print(e.message)
                    else:
                        print(e)
                    print('skipped HDU {}'.format(i))
                meta, data = {}, None

        # Check if an additional frequency axis exists. If so, they should be in the last hdu.
        if hasattr(hdulist[-1].data, 'cfreqs'):
            if verbose:
                print('FITS file contains an additional frequency axis. '
                      'Update the frequency information in cfreqs and cdelts. '
                      'Ignore the original version with equal spacings.')
            meta['ref_cfreqs'] = np.array(hdulist[-1].data['cfreqs'])
            meta['ref_freqdelts'] = np.array(hdulist[-1].data['cdelts'])
        else:
            if verbose:
                print('FITS file does not have an additional frequency axis. '
                      'Use the original version with equal spacings.')
        hdulist.close()

        return meta, data


def write(fname, data, header, mask=None, fix_invalid=True, filled_value=0.0, **kwargs):
    """
    Take a data header pair and write a compressed FITS file.
    Caveat: only 1D, 2D, or 3D images are currently supported by Astropy fits compression.
    To be compressed, the image data array (n-dimensional) must have
    at least n-3 single-dimensional entries.

    Parameters
    ----------
    fname : `str`
        File name, with extension.
    data : `numpy.ndarray`
        n-dimensional data array.
    header : `dict`
        A header dictionary.
    compression_type: `str`, optional
        Compression algorithm: one of 'RICE_1', 'RICE_ONE', 'PLIO_1', 'GZIP_1', 'GZIP_2', 'HCOMPRESS_1'
    hcomp_scale: `float`, optional
        HCOMPRESS scale parameter
    """

    dshape = data.shape
    dim = data.ndim
    if dim - np.count_nonzero(np.array(dshape) == 1) > 3:
        return 0
    else:
        if fix_invalid:
            data[np.isnan(data)] = filled_value
        if kwargs is {}:
            kwargs.update({'compression_type': 'RICE_1', 'quantize_level': 4.0})
        if isinstance(fname, str):
            fname = os.path.expanduser(fname)

        header, data = headersqueeze(header, data)
        hdunew = fits.CompImageHDU(data=data, header=header, **kwargs)
        if mask is None:
            hdulnew = fits.HDUList([fits.PrimaryHDU(), hdunew])
        else:
            hdumask = fits.CompImageHDU(data=mask.astype(np.uint8), **kwargs)
            hdulnew = fits.HDUList([fits.PrimaryHDU(), hdunew, hdumask])
        hdulnew.writeto(fname, output_verify='fix')
        return 1


def wrap(fitsfiles, outfitsfile=None, docompress=False, mask=None, fix_invalid=True, filled_value=0.0, observatory=None,
         imres=None, verbose=False,
         **kwargs):
    '''
    wrap single frequency fits files into a multiple frequencies fits file
    '''
    from astropy.time import Time
    if len(fitsfiles) <= 1:
        print('There is only one files in the fits file list. wrap is aborted!')
        return ''
    else:
        fitsfiles = sorted(fitsfiles)
        nband = len(fitsfiles)
        fits_exist = []
        idx_fits_exist = []
        for sidx, fitsf in enumerate(fitsfiles):
            if os.path.exists(fitsf):
                fits_exist.append(fitsf)
                idx_fits_exist.append(sidx)
        if len(fits_exist) == 0: raise ValueError('None of the input fitsfiles exists!')
        if outfitsfile is None:
            hdu = fits.open(fits_exist[0])
            if observatory is None:
                try:
                    observatory = hdu[-1].header['TELESCOP']
                except:
                    observatory = 'RADIO'
                    print('Failed to acquire telescope information. set as RADIO')
            outfitsfile = Time(hdu[-1].header['DATE-OBS']).strftime('{}.%Y%m%dT%H%M%S.%f.allbd.fits'.format(observatory))
            hdu.close()
        os.system('cp {} {}'.format(fits_exist[0], outfitsfile))
        hdu0 = fits.open(outfitsfile, mode='update')
        header = hdu0[-1].header

        if header['NAXIS'] != 4:
            if verbose:
                print('s1')
            if imres is None:
                cdelts = [325.e8] * nband
                print(
                    'Failed to read bandwidth information. Set as 325 MHz assuming this is EOVSA data. Use the value of keyword CDELT3 with caution.')
            else:
                cdelts = np.squeeze(np.diff(np.array(imres['freq']), axis=1)) * 1e9
            stokesinfo = headerparse(header)
            if stokesinfo['axis'] is None:
                npol = 1
            else:
                npol = int(header['NAXIS{}'.format(stokesinfo['axis'])])
            nbd = nband
            ny = int(header['NAXIS2'])
            nx = int(header['NAXIS1'])

            data = np.zeros((npol, nbd, ny, nx))
            cfreqs = []
            for sidx, fitsf in enumerate(fits_exist):
                hdu = fits.open(fitsf)
                cfreqs.append(hdu[-1].header['RESTFRQ'])
                for pidx in range(npol):
                    if hdu[-1].data.ndim == 2:
                        data[pidx, idx_fits_exist[sidx], :, :] = hdu[-1].data
                    elif hdu[-1].data.ndim == 3:
                        data[pidx, idx_fits_exist[sidx], :, :] = hdu[-1].data[pidx, :, :]
                    else:
                        data[pidx, idx_fits_exist[sidx], :, :] = hdu[-1].data[pidx, 0, :, :]
            cfreqs = np.array(cfreqs)
            cdelts = np.array(cdelts)
            indfreq = np.argsort(cfreqs)
            cfreqs = cfreqs[indfreq]
            cdelts = cdelts[indfreq]
            for pidx in range(npol):
                data[pidx, ...] = data[pidx, indfreq]

            df = np.nanmean(np.diff(cfreqs) / np.diff(idx_fits_exist))  ## in case some of the band is missing
            header['NAXIS'] = 4
            header['NAXIS3'] = nband
            header['CTYPE3'] = 'FREQ'
            header['CRVAL3'] = cfreqs[0]
            header['CDELT3'] = df
            header['CRPIX3'] = 1.0
            header['CUNIT3'] = 'Hz      '
            if stokesinfo['axis'] is None:
                header['NAXIS4'] = 1
                header['CTYPE4'] = 'STOKES'
                header['CRVAL4'] = -5  ## assume XX
                header['CDELT4'] = 1.0
                header['CRPIX4'] = 1.0
                header['CUNIT4'] = '        '
            else:
                for k, v in stokesinfo['headernew'].items():
                    print(k, v)
                    header[k] = v
        else:
            if verbose:
                print('s2')
            npol = int(header['NAXIS4'])
            nbd = nband
            ny = int(header['NAXIS2'])
            nx = int(header['NAXIS1'])

            data = np.zeros((npol, nbd, ny, nx))
            cdelts = []
            cfreqs = []
            for sidx, fitsf in enumerate(fits_exist):
                hdu = fits.open(fitsf)
                cdelts.append(hdu[-1].header['CDELT3'])
                cfreqs.append(hdu[-1].header['CRVAL3'])
                for pidx in range(npol):
                    if hdu[-1].data.ndim == 2:
                        data[pidx, idx_fits_exist[sidx], :, :] = hdu[-1].data
                    else:
                        data[pidx, idx_fits_exist[sidx], :, :] = hdu[-1].data[pidx, 0, :, :]
            cfreqs = np.array(cfreqs)
            cdelts = np.array(cdelts)
            df = np.nanmean(np.diff(cfreqs) / np.diff(idx_fits_exist))  ## in case some of the band is missing
            header['cdelt3'] = df
            header['NAXIS3'] = nband
            header['NAXIS'] = 4
            header['CRVAL3'] = header['CRVAL3'] - df * idx_fits_exist[0]

        for dim1 in range(1, header['NAXIS'] + 1):
            for dim2 in range(1, header['NAXIS'] + 1):
                k = 'PC{:02d}_{:d}'.format(dim1, dim2)
                if dim1 == dim2:
                    header[k] = 1.0
                else:
                    header[k] = 0.0

        if os.path.exists(outfitsfile):
            os.system('rm -rf {}'.format(outfitsfile))

        col1 = fits.Column(name='cfreqs', format='E', array=cfreqs)
        col2 = fits.Column(name='cdelts', format='E', array=cdelts)
        tbhdu = fits.BinTableHDU.from_columns([col1, col2])

        if docompress:
            if fix_invalid:
                data[np.isnan(data)] = filled_value
            if kwargs is {}:
                kwargs.update({'compression_type': 'RICE_1', 'quantize_level': 4.0})
            if isinstance(outfitsfile, str):
                outfitsfile = os.path.expanduser(outfitsfile)

            header, data = headersqueeze(header, data)
            if data.ndim == 4:
                print('only 1D, 2D, or 3D images are currently supported for compression. Aborting compression...')
            else:
                hdunew = fits.CompImageHDU(data=data, header=header, **kwargs)

                if mask is None:
                    hdulnew = fits.HDUList([fits.PrimaryHDU(), hdunew, tbhdu])
                else:
                    hdumask = fits.CompImageHDU(data=mask.astype(np.uint8), **kwargs)
                    hdulnew = fits.HDUList([fits.PrimaryHDU(), hdunew, tbhdu, hdumask])
                hdulnew.writeto(outfitsfile, output_verify='fix')
                return outfitsfile

        hdulnew = fits.HDUList([fits.PrimaryHDU(data=data, header=header), tbhdu])
        hdulnew.writeto(outfitsfile)
        print('wrapped fits written as ' + outfitsfile)
        return outfitsfile



import numpy as np
import os
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from astropy.time import Time
import matplotlib.dates as mdates
from astropy.io import fits
import signalsmooth as ss
import numpy.ma as ma
import matplotlib.cm as cm
import sunpy.map
import astropy.units as u
import pickle
import glob
# from astropy import units as u
# import sunpy.map as smap
from scipy.interpolate import griddata
from scipy.signal import butter, lfilter
from scipy.signal import medfilt
from suncasa.utils import DButil
from scipy.interpolate import splev, splrep
import scipy.ndimage
from IPython import embed
import time
from tqdm import *
from copy import deepcopy
from functools import partial
import multiprocessing as mp
import gc
from matplotlib.widgets import Slider


def butter_bandpass(lowcut, highcut, fs, order=5):
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = butter(order, [low, high], btype='band')
    return b, a


def butter_bandpass_filter(data, lowcut, highcut, fs, order=5):
    b, a = butter_bandpass(lowcut, highcut, fs, order=order)
    y = lfilter(b, a, data)
    return y


def FitSlit(xx, yy, cutwidth, cutang, cutlength, s=None, method='Polyfit', ascending=True):
    if len(xx) <= 3 or method == 'Polyfit':
        '''polynomial fit'''
        out = DButil.polyfit(xx, yy, cutlength, len(xx) - 1 if len(xx) <= 3 else 2)
        xs, ys, posangs = out['xs'], out['ys'], out['posangs']
    else:
        if method == 'Param_Spline':
            '''parametic spline fit'''
            out = DButil.paramspline(xx, yy, cutlength, s=s)
            xs, ys, posangs = out['xs'], out['ys'], out['posangs']
        else:
            '''spline fit'''
            out = DButil.spline(xx, yy, cutlength, s=s)
            xs, ys, posangs = out['xs'], out['ys'], out['posangs']
    if not ascending and (method != 'Param_Spline' or len(xx) <= 3):
        xs, ys = xs[::-1], ys[::-1]
        posangs = posangs[::-1]
    dist = DButil.findDist(xs, ys)
    dists = np.cumsum(dist)
    posangs2 = posangs + np.pi / 2
    cutwidths = dists * np.tan(cutang) + cutwidth
    xs0 = xs - cutwidths / 2. * np.cos(posangs2)
    ys0 = ys - cutwidths / 2. * np.sin(posangs2)
    xs1 = xs + cutwidths / 2. * np.cos(posangs2)
    ys1 = ys + cutwidths / 2. * np.sin(posangs2)
    return {'xcen': xs, 'ycen': ys, 'xs0': xs0, 'ys0': ys0, 'xs1': xs1, 'ys1': ys1, 'cutwidth': cutwidths,
            'posangs': posangs, 'posangs2': posangs2, 'dist': dists}


def MakeSlit(pointDF):
    pointDFtmp = pointDF
    xx = pointDFtmp.loc[:, 'xx'].values
    yy = pointDFtmp.loc[:, 'yy'].values
    if len(pointDFtmp.index) <= 1:
        cutslitplt = {'xcen': [], 'ycen': [], 'xs0': [], 'ys0': [], 'xs1': [], 'ys1': [], 'cutwidth': [],
                      'posangs': [], 'posangs2': [], 'dist': []}
    else:
        # if len(pointDFtmp.index) <= 3:
        cutslitplt = FitSlit(xx, yy, 10, 0.0,
                             200, method='Polyfit')
    return cutslitplt


def getimprofile(data, cutslit, xrange=None, yrange=None):
    num = len(cutslit['xcen'])
    if num > 1:
        intens = np.zeros(num)
        ndy, ndx = data.shape
        if xrange is not None and yrange is not None:
            xs0 = (cutslit['xs0'] - xrange[0]) / (xrange[1] - xrange[0]) * ndx
            xs1 = (cutslit['xs1'] - xrange[0]) / (xrange[1] - xrange[0]) * ndx
            ys0 = (cutslit['ys0'] - yrange[0]) / (yrange[1] - yrange[0]) * ndy
            ys1 = (cutslit['ys1'] - yrange[0]) / (yrange[1] - yrange[0]) * ndy
        else:
            xs0 = cutslit['xs0']
            xs1 = cutslit['xs1']
            ys0 = cutslit['ys0']
            ys1 = cutslit['ys1']
        for ll in range(num):
            inten = DButil.improfile(data, [xs0[ll], xs1[ll]],
                                     [ys0[ll], ys1[ll]], interp='nearest')
            intens[ll] = np.mean(inten)
        intensdist = {'x': cutslit['dist'], 'y': intens}
        return intensdist


def plot_map(smap, dspec=None, diff=False, SymLogNorm=False, linthresh=0.5, returnImAx=False, *args, **kwargs):
    import sunpy.cm.cm as cm  ## to bootstrap sdoaia color map
    import matplotlib.cm as cm
    import matplotlib.colors as colors
    from suncasa.utils import DButil
    clrange = DButil.sdo_aia_scale_dict(wavelength=smap.meta['wavelnth'])
    plt.clf()
    if dspec:
        ax = plt.subplot(121)
    else:
        ax = plt.subplot()
    if 'vmin' in kwargs.keys():
        vmin = kwargs['vmin']
    else:
        vmin = clrange['low']
    if 'vmax' in kwargs.keys():
        vmax = kwargs['vmax']
    else:
        vmax = clrange['high']
    if diff:
        if SymLogNorm:
            norm = colors.SymLogNorm(linthresh=linthresh, vmin=vmin, vmax=vmax)
        else:
            norm = colors.Normalize(vmin=vmin, vmax=vmax)
    else:
        norm = colors.LogNorm(vmin=vmin, vmax=vmax)
    try:
        imshow_args = {'cmap': cm.get_cmap('sdoaia{}'.format(smap.meta['wavelnth'])), 'norm': norm,
                       'interpolation': 'nearest', 'origin': 'lower'}
    except:
        imshow_args = {'cmap': 'gray', 'norm': norm,
                       'interpolation': 'nearest', 'origin': 'lower'}
    if smap.coordinate_system.x == 'HG':
        xlabel = 'Longitude [{lon}]'.format(lon=smap.spatial_units.x)
    else:
        xlabel = 'X-position [{xpos}]'.format(xpos=smap.spatial_units.x)
    if smap.coordinate_system.y == 'HG':
        ylabel = 'Latitude [{lat}]'.format(lat=smap.spatial_units.y)
    else:
        ylabel = 'Y-position [{ypos}]'.format(ypos=smap.spatial_units.y)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    imshow_args.update({'extent': list(smap.xrange.value) + list(smap.yrange.value)})
    if smap.detector == 'HMI':
        im1 = ax.imshow(np.rot90(smap.data, 2), **imshow_args)
    else:
        im1 = ax.imshow(smap.data, **imshow_args)
    plt.title('{} {} {} {}'.format(smap.observatory, smap.detector, smap.wavelength, smap.meta['t_obs']))
    plt.colorbar(im1, ax=ax, label='DN counts per second')
    ax.set_autoscale_on(False)
    if dspec:
        fig = plt.gcf()
        # if figsize:
        #     fig.set_size_inches(figsize)
        # else:
        #     fig.set_size_inches(13, 5)
        ax2 = plt.subplot(122)
        im2 = plt.pcolormesh(dspec['x'], dspec['y'], dspec['dspec'], **dspec['args'])
        date_format = mdates.DateFormatter('%H:%M:%S')
        ax2.xaxis_date()
        ax2.xaxis.set_major_formatter(date_format)
        fig.autofmt_xdate(rotation=30)
        ax2.yaxis.set_label_text(dspec['ytitle'])
        plt.colorbar(im2, ax=ax2, label=dspec['ctitle'])
        ax2.set_autoscale_on(False)
        if 'axvspan' in dspec.keys():
            vspan = ax2.axvspan(dspec['axvspan'][0], dspec['axvspan'][1], alpha=0.5, color='white')
        if 'xs' in dspec.keys() and 'ys' in dspec.keys():
            ax2.plot(dspec['xs'], dspec['ys'], '--', lw=2.0, alpha=0.7, c='black')
        if 'xlim' in dspec.keys():
            ax2.set_xlim(dspec['xlim'])
        if 'ylim' in dspec.keys():
            ax2.set_ylim(dspec['ylim'])
        if returnImAx:
            return ax, im1, ax2, im2, vspan
        else:
            return ax, ax2
    else:
        if returnImAx:
            return ax, im1
        else:
            return ax
            # ax.autoscale(True, 'both', True)
            # ax.autoscale_view(True, True, True)
            # ax.relim(visible_only=True)


def smooth(x, window_len=11, window='hanning'):
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    input:
        x: the input signal
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal

    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)

    see also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter

    TODO: the window parameter could be the window itself if an array instead of a string
    NOTE: length(output) != length(input), to correct this: return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError, "smooth only accepts 1 dimension arrays."

    if x.size < window_len:
        raise ValueError, "Input vector needs to be bigger than window size."

    if window_len < 3:
        return x

    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError, "Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'"

    s = np.r_[x[window_len - 1:0:-1], x, x[-2:-window_len - 1:-1]]
    # print(len(s))
    if window == 'flat':  # moving average
        w = np.ones(window_len, 'd')
    else:
        w = eval('np.' + window + '(window_len)')

    y = np.convolve(w / w.sum(), s, mode='same')
    y = y[window_len - 1:-(window_len - 1)]
    return y


def grid(x, y, z, resX=20, resY=40):
    "Convert 3 column data to matplotlib grid"
    xi = np.linspace(np.nanmin(x), np.nanmax(x), resX)
    yi = np.linspace(np.nanmin(y), np.nanmax(y), resY)
    Z = griddata((x, y), z, (xi[None, :], yi[:, None]), method='cubic')
    X, Y = np.meshgrid(xi, yi)
    return X, Y, Z


def polyfit(x, y, length, deg):
    xs = np.linspace(x.min(), x.max(), length)
    z = np.polyfit(x=x, y=y, deg=deg)
    p = np.poly1d(z)
    ys = p(xs)
    rms = np.sqrt(np.sum((np.polyval(z, x) - y) ** 2) / len(x))
    return {'xs': xs, 'ys': ys, 'rms': rms}


datadir = '/Users/fisher/Desktop/work/2017/LLL/aia/'
mkmc = 1
if mkmc:
    from sunpy.physics.transforms.solar_rotation import mapcube_solar_derotate

    t1, t2 = Time('2014-11-09T10:00:00'), Time('2014-11-09T13:00:00')
    binwdth = 2
    dt_data = 1
    trange = [t1.jd, t2.jd]
    SDOdir = '/Volumes/NAOC-001/work/database/aiaBrowserData/Download/'
    sdofile = DButil.readsdofile(datadir=SDOdir, wavelength='171', jdtime=trange)
    # SDOdir = '/Volumes/NAOC-001/work/2015/20141111/data/prep/0171/'
    # sdofile = glob.glob(SDOdir + 'AIA20141109_*_0171.fits')
    # sdofile = sorted(sdofile)
    # x0, x1, y0, y1 = -660., -575., -200., -115.
    x0, x1, y0, y1 = -850, -500., 730., 1200.  # large
    maplist = []
    print 'Loading fits files....'
    for ll in tqdm(sdofile[::dt_data]):
        maptmp = sunpy.map.Map(ll)
        submaptmp = maptmp.submap(u.Quantity([x0 * u.arcsec, x1 * u.arcsec]),
                                  u.Quantity([y0 * u.arcsec, y1 * u.arcsec]))
        submaptmp = submaptmp.resample(u.Quantity(submaptmp.dimensions) / binwdth)
        if submaptmp.detector == 'HMI':
            pass
        else:
            submaptmp = DButil.normalize_aiamap(submaptmp)
        maplist.append(submaptmp)
        # print '{}/{} loaded.'.format(idx + 1, len(sdofile))
    doderotate = False
    if doderotate:
        mapcube = mapcube_solar_derotate(sunpy.map.Map(maplist, cube=True))
    else:
        mapcube = sunpy.map.Map(maplist, cube=True)
    outfile = '{0}/mapcube_{1}_bin{4}_dtdata{5}_{2}_{3}'.format(datadir, mapcube[0].meta['wavelnth'],
                                                                t1.isot[:-4].replace(':', ''),
                                                                t2.isot[:-4].replace(':', ''), binwdth, dt_data)
    with open(outfile, 'wb') as sf:
        print 'Saving mapcube....'
        pickle.dump(mapcube, sf)
    gc.collect()
else:
    with open('{}/mapcube_{}'.format(datadir, 304), 'rb') as sf:
        print 'Loading mapcube....'
        mapcube = pickle.load(sf)
        # submaplist = []
        # for sidx, smap in enumerate(mapcube0):
        #     submaptmp = smap.submap(u.Quantity([x0 * u.arcsec, x1 * u.arcsec]),
        #                             u.Quantity([y0 * u.arcsec, y1 * u.arcsec]))
        #     submaplist.append(submaptmp)
        # mapcube = mapcube_solar_derotate(sunpy.map.Map(submaplist, cube=True))

rdiff_ratio = 1
mapcubename = 'mapcube_171_bin2_dtdata1_2014-11-09T140000_2014-11-09T200000'
mapcubename = 'mapcube_304_bin2_dtdata1_2014-11-09T150000_2014-11-09T200000'
mapcubename = 'mapcube_171_bin2_dtdata1_2014-11-09T150000_2014-11-09T200000'
mapcubename = 'mapcube_171_bin2_dtdata1_2014-11-09T100000_2014-11-09T130000'
loadmapcube = True
if loadmapcube:
    with open('{}/{}'.format(datadir, mapcubename), 'rb') as sf:
        print 'Loading mapcube ....'
        mapcube = pickle.load(sf)
if mapcube[0].meta['wavelnth'] == 304:
    for idx, smap in enumerate(mapcube):
        mapcube[idx].data[mapcube[idx].data < 1.0] = 1.0

if rdiff_ratio:
    dt_frm = 6
    domedfilt = 0
    # figure()
    # plt.subplot(121)
    # plt.imshow(mapcube[0].data)
    # plt.subplot(122)
    # plt.imshow(medfilt(mapcube[0].data, 3))
    maps_rdiff = []
    datacube = mapcube.as_array()
    if domedfilt:
        for idx, ll in enumerate(tqdm(mapcube)):
            datacube[:, :, idx] = medfilt(datacube[:, :, idx], 3)
    print 'making the rdiff mapcube.....'
    for idx, ll in enumerate(tqdm(mapcube)):
        maps_rdiff.append(deepcopy(ll))
        if idx - dt_frm < 0:
            sidx = 0
        else:
            sidx = idx - dt_frm
        maps_rdiff[idx].data = datacube[:, :, idx] - datacube[:, :, sidx]

    mapcube_rdiff = sunpy.map.Map(maps_rdiff, cube=True)
    outdir = '{0}img_{1}_diff'.format(datadir, mapcubename)
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    else:
        os.system('rm -rf {}/*.png'.format(outdir))
    datacube = mapcube_rdiff.as_array()
    # fig_cutslit = plt.figure(5, figsize=(7, 5))
    # ax = plot_map(mapcube_rdiff[0], vmax=500, vmin=-500, diff=True)
    datacube_ft = np.zeros_like(datacube)
    ny, nx, nt = datacube_ft.shape


    def b_filter(data, lowcut, highcut, fs, ix):
        x = data[ix]
        y = butter_bandpass_filter(x, lowcut, highcut, fs, order=3)
        return {'idx': ix, 'y': y}


    dofilter = True
    if dofilter:
        fs = len(mapcube_rdiff)
        lowcut = 0.1
        highcut = 50
        ncpu = mp.cpu_count() - 1
        print 'filtering the mapcube in time domain.....'
        for ly in tqdm(xrange(ny)):
            b_filter_partial = partial(b_filter, datacube[ly], lowcut, highcut, fs)
            pool = mp.Pool(ncpu)
            res = pool.map(b_filter_partial, xrange(nx))
            pool.close()
            pool.join()
            for lx in xrange(nx):
                datacube_ft[ly, lx] = res[lx]['y']

        for idx, ll in enumerate(tqdm(mapcube_rdiff)):
            mapcube_rdiff[idx].data = datacube_ft[:, :, idx]
        outfile = '{0}/mapcube_{1}_bin{4}_dtdata{5}_{2}_{3}_diff_filter'.format(datadir, mapcube[0].meta['wavelnth'],
                                                                                t1.isot[:-4].replace(':', ''),
                                                                                t2.isot[:-4].replace(':', ''), binwdth,
                                                                                dt_data)
        with open(outfile, 'wb') as sf:
            print 'Saving mapcube....'
            pickle.dump(mapcube_rdiff, sf)
    else:
        mapcube_rdiff_name = 'mapcube_171_bin2_dtdata1_2014-11-09T100000_2014-11-09T130000_diff_filter'
        with open('{}/{}'.format(datadir, mapcube_rdiff_name), 'rb') as sf:
            print 'Loading mapcube ....'
            mapcube_rdiff = pickle.load(sf)
    # mapcube_rdiff = mapcube_rdiff
    mapcube_plot = mapcube_rdiff[30:301]

    # x = datacube[300, 300]
    # y = butter_bandpass_filter(x, lowcut, highcut, fs, order=6)
    # clf()
    # plot(x)
    # plot(y)
    doplot = False
    if doplot:
        fig = plt.figure(1, figsize=(7, 5))
        plt.ioff()
        if mapcube_plot[0].meta['wavelnth'] == 304:
            # vmax, vmin, diff = 2, -2, True
            vmax, vmin, diff = 3, 0.8, False
        elif mapcube_plot[0].meta['wavelnth'] == 193:
            vmax, vmin = 1000, -1000
        elif mapcube_plot[0].meta['wavelnth'] == 171:
            vmax, vmin, diff = 10, -10, True
        elif mapcube_plot[0].meta['wavelnth'] == 335:
            vmax, vmin = 50, -50
        else:
            vmax, vmin = 1000, -1000
        print 'making images.....'
        for smap in tqdm(mapcube_plot):
            ax = plot_map(smap, vmax=vmax, vmin=vmin, diff=diff)
            t_map = Time(smap.meta['t_obs'])
            fig.savefig('{0}/{3}{1}-{2}.png'.format(outdir, smap.meta['wavelnth'],
                                                    t_map.iso.replace(' ', 'T').replace(':',
                                                                                        '').replace('-',
                                                                                                    '')[
                                                    :-4], smap.detector), format='png', dpi=100)
        plt.ion()
else:
    mapcube_plot = deepcopy(mapcube[30:301])
    for idx, smap in enumerate(tqdm(mapcube_plot)):
        smap = DButil.sdo_aia_scale_hdr(smap)
        mapcube_plot[idx].data = smap.data

docutslit = 1
if docutslit:
    class CutslitBuilder:
        def __init__(self, axes, cutwidth=5, cutang=0, cutlength=80):
            self.axes = axes
            self.clickedpoints, = self.axes.plot([], [], 'o', color='white')
            self.slitline, = self.axes.plot([], [], color='white', ls='solid')
            self.slitline0, = self.axes.plot([], [], color='white', ls='dotted')
            self.slitline1, = self.axes.plot([], [], color='white', ls='dotted')
            self.cutlength = cutlength
            self.cutwidth = cutwidth
            self.cutang = cutang
            self.xx = list(self.clickedpoints.get_xdata())
            self.yy = list(self.clickedpoints.get_ydata())
            self.cid = self.clickedpoints.figure.canvas.mpl_connect('button_press_event', self)

        def __call__(self, event):
            # print('button=%d, x=%d, y=%d, xdata=%f, ydata=%f' % (event.button, event.x, event.y, event.xdata, event.ydata))
            if event.inaxes != self.axes:
                return
            if event.button == 1:
                self.xx.append(event.xdata)
                self.yy.append(event.ydata)
            elif event.button == 3:
                self.xx.pop()
                self.yy.pop()
            xx = np.array(self.xx, dtype=np.float64)
            yy = np.array(self.yy, dtype=np.float64)
            self.clickedpoints.set_data(self.xx, self.yy)
            self.clickedpoints.figure.canvas.draw()

            if len(self.xx) <= 1:
                cutslitplt = {'xcen': [], 'ycen': [], 'xs0': [], 'ys0': [], 'xs1': [], 'ys1': [], 'cutwidth': [],
                              'posangs': [], 'posangs2': [], 'dist': []}
            else:
                if len(self.xx) <= 3:
                    cutslitplt = FitSlit(xx, yy, self.cutwidth, self.cutang, self.cutlength, method='Polyfit')
                else:
                    cutslitplt = FitSlit(xx, yy, self.cutwidth, self.cutang, self.cutlength, s=len(xx),
                                         method='Param_Spline')
            self.cutslitplt = cutslitplt
            self.slitline.set_data(cutslitplt['xcen'], cutslitplt['ycen'])
            self.slitline0.set_data(cutslitplt['xs0'], cutslitplt['ys0'])
            self.slitline1.set_data(cutslitplt['xs1'], cutslitplt['ys1'])
            self.slitline.figure.canvas.draw()
            self.slitline0.figure.canvas.draw()
            self.slitline1.figure.canvas.draw()


    dt_frm = 3
    domedfilt = 0
    # mapcubename = 'mapcube_171_bin2_dtdata5_2014-11-09T150000_2014-11-09T170000'

    tplot = []
    for idx, smap in enumerate(tqdm(mapcube_plot)):
        # maps_rdiff.append(deepcopy(smap))
        tplot.append(smap.meta['t_obs'])
    tplot = Time(tplot)
    fig_cutslit = plt.figure(5, figsize=(7, 5))
    clrange = DButil.sdo_aia_scale_dict(mapcube_plot[0].meta['wavelnth'])
    if mapcube_plot[0].meta['wavelnth'] == 304:
        # vmax, vmin, diff = 2, -2, True
        vmax, vmin, diff = 10, 1.0, False
    elif mapcube_plot[0].meta['wavelnth'] == 193:
        vmax, vmin = 1000, -1000
    elif mapcube_plot[0].meta['wavelnth'] == 171:
        vmax, vmin, diff = 10, -10, True
        vmax, vmin, diff = clrange['high'], clrange['low'], False
    elif mapcube_plot[0].meta['wavelnth'] == 335:
        vmax, vmin = 50, -50
    else:
        vmax, vmin = 1000, -1000
    ax, im1 = plot_map(mapcube_plot[0], vmax=vmax, vmin=vmin, diff=diff, returnImAx=True)
    # ax, im1 = plot_map(mapcube_plot[0], vmax=3, vmin=-3, diff=True, returnImAx=True)
    cutslitbuilder = CutslitBuilder(ax, cutwidth=20, cutang=np.pi / 60, cutlength=150)
    axcolor = 'lightgoldenrodyellow'
    axframe = plt.axes([0.25, 0.05, 0.65, 0.03], facecolor=axcolor)
    sframe = Slider(axframe, 'frame', 0, len(mapcube_plot) - 1, valinit=0, valfmt='%0.0f')


    def update(val):
        frm = sframe.val
        smap = mapcube_plot[int(frm)]
        # smap.data[smap.data<1]=1
        im1.set_data(smap.data)
        ax.set_title('{} {} {} {}'.format(smap.observatory, smap.detector, smap.wavelength, smap.meta['t_obs']))
        fig_cutslit.canvas.draw()


    sframe.on_changed(update)

    to_save = 0
    to_restore = 1
    if to_save:
        cutslit = {'x': cutslitbuilder.clickedpoints.get_xdata(), 'y': cutslitbuilder.clickedpoints.get_ydata(),
                   'cutslit': cutslitbuilder.cutslitplt}
        with open('{}cutslit2'.format(datadir), 'wb') as sf:
            pickle.dump(cutslit, sf)
    if to_restore:
        with open('{}cutslit2'.format(datadir), 'rb') as sf:
            cutslit = pickle.load(sf)
        cutslitbuilder.clickedpoints.set_data(cutslit['x'], cutslit['y'])
        cutslitbuilder.clickedpoints.figure.canvas.draw()
        cutslitbuilder.cutslitplt = cutslit['cutslit']
        cutslitbuilder.slitline.set_data(cutslit['cutslit']['xcen'], cutslit['cutslit']['ycen'])
        cutslitbuilder.slitline0.set_data(cutslit['cutslit']['xs0'], cutslit['cutslit']['ys0'])
        cutslitbuilder.slitline1.set_data(cutslit['cutslit']['xs1'], cutslit['cutslit']['ys1'])
        cutslitbuilder.slitline.figure.canvas.draw()
        cutslitbuilder.slitline0.figure.canvas.draw()
        cutslitbuilder.slitline1.figure.canvas.draw()
    dostackplot = 1
    if dostackplot:
        stackplt = []
        print 'making the stack plot...'
        for idx, smap in enumerate(tqdm(mapcube_plot)):
            intens = getimprofile(smap.data, cutslitbuilder.cutslitplt,
                                  xrange=smap.xrange.value,
                                  yrange=smap.yrange.value)
            stackplt.append(intens['y'])
        if len(stackplt) > 1:
            stackplt = np.vstack(stackplt)
            stackplt = stackplt.transpose()
        # if mapcube_plot[0].meta['wavelnth'] == 304:
        #     # vmax, vmin, diff = 2, -2, True
        #     vmax, vmin, diff = 3, 0.8, False
        # elif mapcube_plot[0].meta['wavelnth'] == 193:
        #     vmax, vmin = 1000, -1000
        # elif mapcube_plot[0].meta['wavelnth'] == 171:
        #     vmax, vmin, diff = 5, -5, True
        # elif mapcube_plot[0].meta['wavelnth'] == 335:
        #     vmax, vmin = 50, -50
        # else:
        #     vmax, vmin = 1000, -1000
        norm = colors.Normalize(vmin=vmin, vmax=vmax)
        # norm = colors.LogNorm(vmin=vmin, vmax=vmax)
        cutslitplt = cutslitbuilder.cutslitplt
        dspec = {'dspec': stackplt, 'x': tplot.plot_date, 'y': cutslitplt['dist'],
                 'ytitle': 'Distance [arcsec]',
                 'ctitle': 'DN counts per second',
                 'args': {'norm': norm, 'cmap': cm.get_cmap('sdoaia{}'.format(mapcube_plot[0].meta['wavelnth']))}}
        with open(outfile + '_dspec', 'wb') as sf:
            print 'Saving dspec....'
            pickle.dump(dspec, sf)

        with open(outfile + '_dspec', 'rb') as sf:
            print 'Loading dspec....'
            dspec = pickle.load(sf)

        fig_stackplt = plt.figure(6, figsize=(12, 5))
        dtplot = np.mean(np.diff(tplot.plot_date))
        dspec['axvspan'] = [tplot[0].plot_date, tplot[0].plot_date + dtplot]
        ax, im1, ax2, im2, vspan = plot_map(mapcube_plot[0], dspec, vmax=vmax, vmin=vmin, diff=diff,
                                            returnImAx=True)
        ax.plot(cutslitplt['xcen'], cutslitplt['ycen'], color='white', ls='solid')
        ax.plot(cutslitplt['xs0'], cutslitplt['ys0'], color='white', ls='dotted')
        ax.plot(cutslitplt['xs1'], cutslitplt['ys1'], color='white', ls='dotted')
        axcolor = 'lightgoldenrodyellow'
        axframe2 = plt.axes([0.25, 0.05, 0.65, 0.03], facecolor=axcolor)
        sframe2 = Slider(axframe2, 'frame', 0, len(mapcube_plot) - 1, valinit=0, valfmt='%0.0f')


        def update2(val):
            frm = int(sframe2.val)
            smap = mapcube_plot[frm]
            im1.set_data(smap.data)
            ax.set_title('{} {} {} {}'.format(smap.observatory, smap.detector, smap.wavelength, smap.meta['t_obs']))
            vspan_xy = vspan.get_xy()
            vspan_xy[np.array([0, 1, 4]), 0] = tplot[frm].plot_date
            vspan_xy[np.array([2, 3]), 0] = tplot[frm].plot_date + dtplot
            vspan.set_xy(vspan_xy)
            fig_stackplt.canvas.draw()


        sframe2.on_changed(update2)

        doplot = False
        if doplot:
            outdir = '{0}img_{1}_diff_stackplt'.format(datadir, mapcubename)
            if not os.path.exists(outdir):
                os.makedirs(outdir)
            else:
                os.system('rm -rf {}/*.png'.format(outdir))
            print 'producing images...'
            for xidx, smap in enumerate(tqdm(mapcube_plot)):
                dspec['axvspan'] = [tplot[xidx].plot_date, tplot[xidx].plot_date + dtplot]
                ax, ax2 = plot_map(smap, dspec, vmax=vmax, vmin=vmin, diff=True)
                ax.plot(cutslitplt['xcen'], cutslitplt['ycen'], color='white', ls='solid')
                ax.plot(cutslitplt['xs0'], cutslitplt['ys0'], color='white', ls='dotted')
                ax.plot(cutslitplt['xs1'], cutslitplt['ys1'], color='white', ls='dotted')
                fig_stackplt.savefig('{0}/{3}{1}-{2}.png'.format(outdir, smap.meta['wavelnth'],
                                                                 tplot[xidx].iso.replace(' ', 'T').replace(':',
                                                                                                           '').replace(
                                                                     '-',
                                                                     '')[
                                                                 :-4], smap.detector), format='png', dpi=100)



import os
import numpy as np
from taskinit import tb, casalog
from concat_cli import concat_cli as concat
from clearcal_cli import clearcal_cli as clearcal
from split_cli import split_cli as split


def concateovsa(vis, concatvis, datacolumn='corrected', keep_orig_ms=True, cols2rm="model,corrected", freqtol="", dirtol="", respectname=False,
                timesort=True, copypointing=True, visweightscale=[], forcesingleephemfield=""):
    if concatvis[-1] == os.path.sep:
        concatvis = concatvis[:-1]
    if os.path.sep not in concatvis:
        visprefix = './'
    else:
        visprefix = os.path.dirname(concatvis) + os.path.sep
    msfiles = vis
    msfiles_ = []
    for idx, ll in enumerate(msfiles):
        if str(ll).endswith('/'):
            msfiles[idx] = str(ll)[:-1]
    datacolumn = datacolumn.lower()
    if datacolumn == 'data':
        print('DATA columns will be concatenated.')
        for ll in msfiles:
            clearcal(vis=str(ll), addmodel=True)
    elif datacolumn == 'corrected':
        # try:
        print('CORRECTED columns will be concatenated.')
        tmpdir = os.path.join(visprefix, 'tmp_ms') + os.path.sep
        if not os.path.exists(tmpdir):
            os.makedirs(tmpdir)
        for ll in msfiles:
            msfile_ = os.path.join(tmpdir, os.path.basename(str(ll)))
            msfiles_.append(msfile_)
            split(vis=str(ll), outputvis=msfile_, datacolumn='corrected')
            clearcal(vis=msfile_, addmodel=True)
    else:
        raise ValueError('Please set datacolumn to be "data" or "corrected"!')

    if msfiles_:
        concat(vis=msfiles_, concatvis=concatvis, freqtol=freqtol, dirtol=dirtol, respectname=respectname, timesort=timesort,
               copypointing=copypointing, visweightscale=visweightscale, forcesingleephemfield=forcesingleephemfield)
        os.system('rm -rf {}'.format(tmpdir))
    else:
        concat(vis=msfiles, concatvis=concatvis, freqtol=freqtol, dirtol=dirtol, respectname=respectname, timesort=timesort,
               copypointing=copypointing, visweightscale=visweightscale, forcesingleephemfield=forcesingleephemfield)
    # Change all observation ids to be the same (zero)
    tb.open(concatvis + '/OBSERVATION', nomodify=False)
    nobs = tb.nrows()
    tim0 = tb.getcell('TIME_RANGE', 0)[0]
    tim1 = tb.getcell('TIME_RANGE', nobs - 1)[1]
    tb.removerows([i + 1 for i in range(nobs - 1)])
    tb.putcell('TIME_RANGE', 0, [tim0, tim1])
    tb.close()

    tb.open(concatvis + '/DATA_DESCRIPTION', nomodify=False)
    nrows = tb.nrows()
    pol_id = tb.getcol('POLARIZATION_ID')
    tb.removerows(np.where(pol_id != 0)[0])
    tb.close()

    tb.open(concatvis, nomodify=False)
    dd_id = tb.getcol('DATA_DESC_ID')
    idx_dd_id, = np.where(dd_id >= nrows / 2)
    dd_id[idx_dd_id] = dd_id[idx_dd_id] - nrows / 2
    tb.putcol('DATA_DESC_ID', dd_id)
    tb.close()

    tb.open(concatvis + '/FIELD', nomodify=False)
    nobs = tb.nrows()
    tb.removerows([i + 1 for i in range(nobs - 1)])
    tb.close()

    tb.open(concatvis + '/SOURCE', nomodify=False)
    nobs = tb.nrows()
    tb.removerows([i + 1 for i in range(nobs - 1)])
    tb.close()

    tb.open(concatvis, nomodify=False)
    obsid = tb.getcol('OBSERVATION_ID')
    newobsid = np.zeros(len(obsid), dtype='int')
    tb.putcol('OBSERVATION_ID', newobsid)
    fldid = tb.getcol('FIELD_ID')
    newfldid = np.zeros(len(fldid), dtype='int')
    tb.putcol('FIELD_ID', newfldid)
    colnames = tb.colnames()

    cols2rm = cols2rm.upper()
    cols2rm = cols2rm.split(',')
    for l in range(len(cols2rm)):
        col = cols2rm[l] + '_DATA'
        if col in colnames:
            try:
                tb.removecols(col)
                print('Column {} removed.'.format(col))
            except:
                pass
    tb.close()

    if msfiles_ != [] and msfiles_ != msfiles:
        for ll in msfiles_:
            os.system('rm -rf {}'.format(ll))
    if not keep_orig_ms:
        for ll in msfiles:
            os.system('rm -rf {}'.format(ll))
            os.system('rm -rf {}.flagversions'.format(ll))



import os
import numpy as np
from datetime import datetime
from math import *
# import jdutil
import bisect
import pdb
from casatools import ms as mstool
from casatools import table, quanta, image
from astropy.time import Time
from sunpy import sun, coordinates
import astropy.units as u
import warnings
#from suncasa.utils 
import fitsutils as fu

ms = mstool()
tb = table()
qa = quanta()
ia = image()

try:
    from astropy.io import fits as pyfits
except:
    try:
        import pyfits
    except ImportError:
        raise ImportError('Neither astropy nor pyfits exists in this CASA installation')


def ms_clearhistory(msfile):
    from taskinit import tb
    tb_history = msfile + '/HISTORY'
    os.system('cp -r {0} {0}_bk'.format(tb_history))
    tb.open(tb_history, nomodify=False)
    nrows = tb.nrows()
    if nrows > 0:
        tb.removerows(range(nrows))
    tb.close()


def ms_restorehistory(msfile):
    tb_history = msfile + '/HISTORY'
    os.system('mv {0}_bk {0}'.format(tb_history))


def read_horizons(t0=None, dur=None, vis=None, observatory=None, verbose=False):
    '''
    This function visits JPL Horizons to retrieve J2000 topocentric RA and DEC of the solar disk center
    as a function of time.

    Keyword arguments:
    t0: Referece time in astropy.Time format
    dur: duration of the returned coordinates. Default to 2 minutes
    vis: CASA visibility dataset (in measurement set format). If provided, use entire duration from
         the visibility data
    observatory: observatory code (from JPL Horizons). If not provided, use information from visibility.
         if no visibility found, use earth center (code=500)
    verbose: True to provide extra information

    Usage:
    >>> from astropy.time import Time
    >>> out = read_horizons(t0=Time('2017-09-10 16:00:00'), observatory='-81')
    >>> out = read_horizons(vis = 'mydata.ms')

    History:
    BC (sometime in 2014): function was first wrote, followed by a number of edits by BC and SY
    BC (2019-07-16): Added docstring documentation

    '''
    import urllib.request as urllib2
    import ssl
    if not t0 and not vis:
        t0 = Time.now()
    if not dur:
        dur = 1. / 60. / 24.  # default to 2 minutes
    if t0:
        try:
            btime = Time(t0)
        except:
            print('input time ' + str(t0) + ' not recognized')
            return -1
    if vis:
        if not os.path.exists(vis):
            print('Input ms data ' + vis + ' does not exist! ')
            return -1
        try:
            # ms.open(vis)
            # summary = ms.summary()
            # ms.close()
            # btime = Time(summary['BeginTime'], format='mjd')
            # etime = Time(summary['EndTime'], format='mjd')
            ## alternative way to avoid conflicts with importeovsa, if needed -- more time consuming
            if observatory == 'geocentric':
                observatory = '500'
            else:
                ms.open(vis)
                metadata = ms.metadata()
                if metadata.observatorynames()[0] == 'EVLA':
                    observatory = '-5'
                elif metadata.observatorynames()[0] == 'EOVSA':
                    observatory = '-81'
                elif metadata.observatorynames()[0] == 'ALMA':
                    observatory = '-7'
                ms.close()
            tb.open(vis)
            btime_vis = Time(tb.getcell('TIME', 0) / 24. / 3600., format='mjd')
            etime_vis = Time(tb.getcell('TIME', tb.nrows() - 1) / 24. / 3600., format='mjd')
            tb.close()
            if verbose:
                print("Beginning time of this scan " + btime_vis.iso)
                print("End time of this scan " + etime_vis.iso)

            # extend the start and end time for jpl horizons by 0.5 hr on each end
            btime = Time(btime_vis.mjd - 0.5 / 24., format='mjd')
            dur = etime_vis.mjd - btime_vis.mjd + 1.0 / 24.
        except:
            print('error in reading ms file: ' + vis + ' to obtain the ephemeris!')
            return -1

    # default the observatory to geocentric, if none provided
    if not observatory:
        observatory = '500'

    etime = Time(btime.mjd + dur, format='mjd')

    try:
        cmdstr = "https://ssd.jpl.nasa.gov/horizons_batch.cgi?batch=1&TABLE_TYPE='OBSERVER'&QUANTITIES='1,17,20'&CSV_FORMAT='YES'&ANG_FORMAT='DEG'&CAL_FORMAT='BOTH'&SOLAR_ELONG='0,180'&CENTER='{}@399'&COMMAND='10'&START_TIME='".format(
            observatory) + btime.iso.replace(' ', ',') + "'&STOP_TIME='" + etime.iso[:-4].replace(' ',
                                                                                                  ',') + "'&STEP_SIZE='1m'&SKIP_DAYLT='NO'&EXTRA_PREC='YES'&APPARENT='REFRACTED'"
        cmdstr = cmdstr.replace("'", "%27")
        try:
            context = ssl._create_unverified_context()
            f = urllib2.urlopen(cmdstr, context=context)
        except:
            f = urllib2.urlopen(cmdstr)
        lines = f.readlines()
        f.close()
    except:
        # todo use geocentric coordinate for the new VLA data
        import requests, collections
        params = collections.OrderedDict()
        params['batch'] = '1'
        params['TABLE_TYPE'] = "'OBSERVER'"
        params['QUANTITIES'] = "'1,17,20'"
        params['CSV_FORMAT'] = "'YES'"
        params['ANG_FORMAT'] = "'DEG'"
        params['CAL_FORMAT'] = "'BOTH'"
        params['SOLAR_ELONG'] = "'0,180'"
        if observatory == '500':
            params['CENTER'] = "'500'"
        else:
            params['CENTER'] = "'{}@399'".format(observatory)
        params['COMMAND'] = "'10'"
        params['START_TIME'] = "'{}'".format(btime.iso[:-4].replace(' ', ','))
        params['STOP_TIME'] = "'{}'".format(etime.iso[:-4].replace(' ', ','))
        params['STEP_SIZE'] = "'1m'"
        params['SKIP_DAYLT'] = "'NO'"
        params['EXTRA_PREC'] = "'YES'"
        params['APPAENT'] = "'REFRACTED'"
        results = requests.get("https://ssd.jpl.nasa.gov/horizons_batch.cgi", params=params)
        lines = [ll for ll in results.iter_lines()]

    nline = len(lines)
    istart = 0
    for i in range(nline):
        line = lines[i].decode('utf-8')
        if line[0:5] == '$$SOE':  # start recording
            istart = i + 1
        if line[0:5] == '$$EOE':  # end recording
            iend = i
    newlines = lines[istart:iend]
    nrec = len(newlines)
    ephem_ = []
    t = []
    ra = []
    dec = []
    p0 = []
    delta = []
    for line in newlines:
        items = line.decode('utf-8').split(',')
        t.append(Time(float(items[1]), format='jd').mjd)
        ra.append(np.radians(float(items[4])))
        dec.append(np.radians(float(items[5])))
        p0.append(float(items[6]))
        delta.append(float(items[8]))
    # convert list of dictionary to a dictionary of arrays
    ephem = {'time': t, 'ra': ra, 'dec': dec, 'p0': p0, 'delta': delta}
    return ephem


def read_msinfo(vis=None, msinfofile=None, use_scan_time=True):
    import glob
    # read MS information #
    msinfo = dict.fromkeys(['vis', 'scans', 'fieldids', 'btimes', 'btimestr', 'inttimes', 'ras', 'decs', 'observatory'])
    ms.open(vis)
    metadata = ms.metadata()
    observatory = metadata.observatorynames()[0]
    scans = ms.getscansummary()
    scanids = sorted(scans.keys(), key=lambda x: int(x))
    nscanid = len(scanids)
    btimes = []
    btimestr = []
    etimes = []
    fieldids = []
    inttimes = []
    dirs = []
    ras = []
    decs = []
    ephem_file = glob.glob(vis + '/FIELD/EPHEM*SUN.tab')
    if ephem_file:
        print('Loading ephemeris info from {}'.format(ephem_file[0]))
        tb.open(ephem_file[0])
        col_ra = tb.getcol('RA')
        col_dec = tb.getcol('DEC')
        col_mjd = tb.getcol('MJD')
        if use_scan_time:
            from scipy.interpolate import interp1d
            f_ra = interp1d(col_mjd, col_ra)
            f_dec = interp1d(col_mjd, col_dec)
            for idx, scanid in enumerate(scanids):
                btimes.append(scans[scanid]['0']['BeginTime'])
                etimes.append(scans[scanid]['0']['EndTime'])
                fieldid = scans[scanid]['0']['FieldId']
                fieldids.append(fieldid)
                inttimes.append(scans[scanid]['0']['IntegrationTime'])
            ras = f_ra(np.array(btimes))
            decs = f_dec(np.array(btimes))
            ras = qa.convert(qa.quantity(ras, 'deg'), 'rad')
            decs = qa.convert(qa.quantity(decs, 'deg'), 'rad')
        else:
            ras = qa.convert(qa.quantity(col_ra, 'deg'), 'rad')
            decs = qa.convert(qa.quantity(col_dec, 'deg'), 'rad')
    else:
        for idx, scanid in enumerate(scanids):
            btimes.append(scans[scanid]['0']['BeginTime'])
            etimes.append(scans[scanid]['0']['EndTime'])
            fieldid = scans[scanid]['0']['FieldId']
            fieldids.append(fieldid)
            inttimes.append(scans[scanid]['0']['IntegrationTime'])
            dir = ms.getfielddirmeas('PHASE_DIR', fieldid)
            dirs.append(dir)
            ras.append(dir['m0'])
            decs.append(dir['m1'])
    ms.close()
    btimestr = [qa.time(qa.quantity(btimes[idx], 'd'), form='fits', prec=10)[0] for idx in range(nscanid)]
    msinfo['vis'] = vis
    msinfo['scans'] = scans
    msinfo['fieldids'] = fieldids
    msinfo['btimes'] = btimes
    msinfo['btimestr'] = btimestr
    msinfo['inttimes'] = inttimes
    msinfo['ras'] = ras
    msinfo['decs'] = decs
    msinfo['observatory'] = observatory
    if msinfofile:
        np.savez(msinfofile, vis=vis, scans=scans, fieldids=fieldids, btimes=btimes, btimestr=btimestr,
                 inttimes=inttimes, ras=ras, decs=decs,
                 observatory=observatory)
    return msinfo


def ephem_to_helio(vis=None, ephem=None, msinfo=None, reftime=None, polyfit=None, usephacenter=False):
    '''1. Take a solar ms database, read the scan and field information, find out the pointings (in RA and DEC)
       2. Compare with the ephemeris of the solar disk center (in RA and DEC)
       3. Generate VLA pointings in heliocentric coordinates
       inputs:
           msinfo: CASA MS information, output from read_msinfo
           ephem: solar ephem, output from read_horizons
           reftime: list of reference times (e.g., used for imaging)
                    CASA standard time format, either a single time (e.g., '2012/03/03/12:00:00'
                    or a time range (e.g., '2012/03/03/12:00:00~2012/03/03/13:00:00'. If the latter,
                    take the midpoint of the timerange for reference. If no date specified, take
                    the date of the first scan
           polyfit: ONLY works for MS database with only one source with continously tracking;
                    not recommanded unless scan length is too long and want to have very high accuracy
           usephacenter: Bool -- if True, correct for the RA and DEC in the ms file based on solar empheris.
                                 Otherwise assume the phasecenter is correctly pointed to the solar disk center
                                 (EOVSA case)

       return value:
           helio: a list of VLA pointing information
                   reftimestr: reference time, in FITS format string
                   reftime: reference time, in mjd format
                   ra: actual RA of phasecenter in the ms file at the reference time (interpolated)
                   dec: actual DEC of phasecenter in the ms file at the reference time (interpolated)
                   # CASA uses only RA and DEC of the closest field (e.g. in clean) #
                   ra_fld: right ascention of the CASA reference pointing direction
                   dec_fld: declination of the CASA reference pointing direction
                   raoff: RA offset of the phasecenter in the ms file to solar center
                   decoff: DEC offset of the phasecenter in the ms file to solar center
                   refx: heliocentric X offset of the phasecenter in the ms file to solar center
                   refy: heliocentric Y offset of the phasecenter in the ms file to solar center
    ######## Example #########
         msfile='sun_C_20140910T221952-222952.10s.cal.ms'
         ephemfile='horizons_sun_20140910.radecp'
         ephem=vla_prep.read_horizons(ephemfile=ephemfile)
         msinfo=vla_prep.read_msinfo(msfile=msfile)
         polyfit=0
         reftime = '22:25:20~22:25:40'
    '''
    if not vis or not os.path.exists(vis):
        raise ValueError('Please provide information of the MS database!')
    if not ephem:
        ephem = read_horizons(vis=vis)
    if not msinfo:
        msinfo0 = read_msinfo(vis)
    else:
        if isinstance(msinfo, str):
            try:
                msinfo0 = np.load(msinfo)
            except:
                raise ValueError('The specified input msinfo file does not exist!')
        elif isinstance(msinfo, dict):
            msinfo0 = msinfo
        else:
            raise ValueError('msinfo should be either a numpy npz or a dictionary')
    print('msinfo is derived from: {0:s}'.format(msinfo0['vis']))
    scans = msinfo0['scans']
    fieldids = msinfo0['fieldids']
    btimes = msinfo0['btimes']
    inttimes = msinfo0['inttimes']
    ras = msinfo0['ras']
    decs = msinfo0['decs']
    if 'observatory' in msinfo0.keys():
        if msinfo0['observatory'] == 'EOVSA':
            usephacenter = False
    if type(ras) is list:
        ra_rads = [ra['value'] for ra in ras]
    elif type(ras) is dict:
        ra_rads = ras['value']
    else:
        print('Type of msinfo0["ras"] unrecognized.')
        return 0
    if type(decs) is list:
        dec_rads = [dec['value'] for dec in decs]
    elif type(decs) is dict:
        dec_rads = decs['value']
    else:
        print('Type of msinfo0["decs"] unrecognized.')
    # fit 2nd order polynomial fits to the RAs and DECs #
    if polyfit:
        cra = np.polyfit(btimes, ra_rads, 2)
        cdec = np.polyfit(btimes, dec_rads, 2)

    # find out phase center infomation in ms according to the input time or timerange #
    if not reftime:
        raise ValueError('Please specify a reference time of the image')
    if type(reftime) == str:
        reftime = [reftime]
    if (not isinstance(reftime, list)):
        print('input "reftime" is not a valid list. Abort...')

    nreftime = len(reftime)
    helio = []
    for reftime0 in reftime:
        helio0 = dict.fromkeys(
            ['reftimestr', 'reftime', 'ra', 'dec', 'ra_fld', 'dec_fld', 'raoff', 'decoff', 'refx', 'refy', 'p0'])
        helio0['reftimestr'] = reftime0
        if '~' in reftime0:
            # if reftime0 is specified as a timerange
            try:
                [tbg0, tend0] = reftime0.split('~')
                tbg_d = qa.getvalue(qa.convert(qa.totime(tbg0), 'd'))[0]
                tend_d = qa.getvalue(qa.convert(qa.totime(tend0), 'd'))[0]
                tdur_s = (tend_d - tbg_d) * 3600. * 24.
                # if no date is specified, add up the date of the first scan
                if tend_d < 1.:
                    if tend_d >= tbg_d:
                        tend_d += int(btimes[0])
                    else:
                        tend_d += int(btimes[0]) + 1
                if tbg_d < 1.:
                    tbg_d += int(btimes[0])
                tref_d = (tbg_d + tend_d) / 2.
            except:
                print('Error in converting the input reftime: ' + str(reftime0) + '. Aborting...')
        else:
            # if reftime0 is specified as a single value
            try:
                tref_d = qa.getvalue(qa.convert(qa.totime(reftime0), 'd'))[0]
                # if no date is specified, add up the date of the first scan
                if tref_d < 1.:
                    tref_d += int(btimes[0])
                tbg_d = tref_d
                # use the intergration time
                # ind = bisect.bisect_left(btimes, tref_d)
                # if msinfo0['etimes']:
                #    tdur_s = inttimes[ind - 1]
                # else:
                #    tdur_s = np.mean(inttimes)
                tdur_s = 1.
            except:
                print('Error in converting the input reftime: ' + str(reftime0) + '. Aborting...')
        helio0['reftime'] = tref_d
        # helio0['date-obs'] = qa.time(qa.quantity(tbg_d, 'd'), form='fits', prec=10)[0]
        # helio0['exptime'] = tdur_s

        # find out phase center RA and DEC in the measurement set according to the reference time
        # if polyfit, then use the 2nd order polynomial coeffs
        inttime = np.nanmean(inttimes)
        ind = bisect.bisect_left(btimes, tref_d)
        if ind > 1:
            dt = tref_d - btimes[ind - 1]
            if ind < len(btimes):
                scanlen = btimes[ind] - btimes[ind - 1]
                (ra_b, ra_e) = (ra_rads[ind - 1], ra_rads[ind])
                (dec_b, dec_e) = (dec_rads[ind - 1], dec_rads[ind])
            if ind >= len(btimes):
                scanlen = btimes[ind - 1] - btimes[ind - 2]
                (ra_b, ra_e) = (ra_rads[ind - 2], ra_rads[ind - 1])
                (dec_b, dec_e) = (dec_rads[ind - 2], dec_rads[ind - 1])
        if ind == 1:  # only one scan exists (e.g., imported from AIPS)
            ra_b = ra_rads[ind - 1]
            ra_e = ra_b
            dec_b = dec_rads[ind - 1]
            dec_e = dec_b
            scanlen = 10.  # radom value
            dt = 0.
        if ind < 1:
            if np.abs((tref_d - btimes[0]) * 24 * 3600) < inttime / 2.0:
                ind = 1
                ra_b = ra_rads[ind - 1]
                ra_e = ra_b
                dec_b = dec_rads[ind - 1]
                dec_e = dec_b
                scanlen = 10.  # radom value
                dt = 0.
            else:
                raise ValueError('Reference time does not fall into the scan list!')
        if polyfit:
            ra = cra[0] * tref_d ** 2. + cra[1] * tref_d + cra[2]
            dec = cdec[0] * tref_d ** 2. + cdec[1] * tref_d + cdec[2]
        # if not, use linearly interpolated RA and DEC at the beginning of this scan and next scan
        else:
            ra = ra_b + (ra_e - ra_b) / scanlen * dt
            dec = dec_b + (dec_e - dec_b) / scanlen * dt
        if ra < 0:
            ra += 2. * np.pi
        if ra_b < 0:
            ra_b += 2. * np.pi

        # compare with ephemeris from JPL Horizons
        time0 = ephem['time']
        ra0 = ephem['ra']
        dec0 = ephem['dec']
        p0 = ephem['p0']
        delta0 = ephem['delta']
        if len(time0) > 1:
            ind = bisect.bisect_left(time0, tref_d)
            dt0 = time0[ind] - time0[ind - 1]
            dt_ref = tref_d - time0[ind - 1]
            dra0 = ra0[ind] - ra0[ind - 1]
            ddec0 = dec0[ind] - dec0[ind - 1]
            dp0 = p0[ind] - p0[ind - 1]
            ddelta0 = delta0[ind] - delta0[ind - 1]
            ra0 = ra0[ind - 1] + dra0 / dt0 * dt_ref
            dec0 = dec0[ind - 1] + ddec0 / dt0 * dt_ref
            p0 = p0[ind - 1] + dp0 / dt0 * dt_ref
            delta0 = delta0[ind - 1] + ddelta0 / dt0 * dt_ref
        else:
            try:
                ra0 = ra0[0]
                dec0 = dec0[0]
                p0 = p0[0]
                delta0 = delta0[0]
            except:
                print("Error in retrieving info from ephemeris!")
        if ra0 < 0:
            ra0 += 2. * np.pi

        # RA and DEC offset in arcseconds
        decoff = degrees((dec - dec0)) * 3600.
        raoff = degrees((ra - ra0) * cos(dec)) * 3600.
        # Convert into heliocentric offsets
        prad = -radians(p0)
        refx = (-raoff) * cos(prad) - decoff * sin(prad)
        refy = (-raoff) * sin(prad) + decoff * cos(prad)
        helio0['ra'] = ra  # ra of the actual pointing
        helio0['dec'] = dec  # dec of the actual pointing
        helio0['ra_fld'] = ra_b  # ra of the field, used as the reference in e.g., clean
        helio0['dec_fld'] = dec_b  # dec of the field, used as the refenrence in e.g., clean
        helio0['raoff'] = raoff
        helio0['decoff'] = decoff
        if usephacenter:
            helio0['refx'] = refx
            helio0['refy'] = refy
        else:
            helio0['refx'] = 0.
            helio0['refy'] = 0.
        helio0['p0'] = p0
        # helio['r_sun']=np.degrees(R_sun.value/(au.value*delta0))*3600. #in arcsecs
        helio.append(helio0)
    return helio


def getbeam(imagefile=None, beamfile=None):
    ia = iatool()
    if not imagefile:
        raise ValueError('Please specify input images')
    bmaj = []
    bmin = []
    bpa = []
    beamunit = []
    bpaunit = []
    chans = []
    nimg = len(imagefile)
    for n in range(nimg):
        img = imagefile[n]
        bmaj_ = []
        bmin_ = []
        bpa_ = []
        if not os.path.exists(img):
            warnings.warn('{} does not exist!'.format(img))
            beamunit_ = []
            bpaunit_ = []
            chans_ = []
        else:
            ia.open(img)
            sum = ia.summary()
            ia.close()
            ia.done()
            if sum.has_key('perplanebeams'):  # beam vary with frequency
                nbeams = sum['perplanebeams']['nChannels']
                beams = sum['perplanebeams']['beams']
                chans_ = [key[1:] for key in beams.keys()]
                chans_.sort(key=float)
                for chan in chans_:
                    bmaj0 = beams['*' + chan]['*0']['major']['value']
                    bmaj_.append(bmaj0)
                    bmin0 = beams['*' + chan]['*0']['minor']['value']
                    bmin_.append(bmin0)
                    bpa0 = beams['*' + chan]['*0']['positionangle']['value']
                    bpa_.append(bpa0)
                beamunit_ = beams['*' + chans_[0]]['*0']['major']['unit']
                bpaunit_ = beams['*' + chans_[0]]['*0']['positionangle']['unit']
            if sum.has_key('restoringbeam'):  # only one beam
                bmaj_.append(sum['restoringbeam']['major']['value'])
                bmin_.append(sum['restoringbeam']['minor']['value'])
                bpa_.append(sum['restoringbeam']['positionangle']['value'])
                beamunit_ = sum['restoringbeam']['major']['unit']
                bpaunit_ = sum['restoringbeam']['positionangle']['unit']
                nbeams = 1
                chans_ = [0]

        bmaj.append(bmaj_)
        bmin.append(bmin_)
        bpa.append(bpa_)
        beamunit.append(beamunit_)
        bpaunit.append(bpaunit_)
        chans.append(chans_)
    if beamfile:  # write beams to ascii file
        print('Writing beam info to ascii file...')
        f = open(beamfile, 'w')
        f.write('CHANNEL No., BMAJ (' + beamunit[0] + '), BMIN (' + beamunit[0] + '), BPA (' + bpaunit[0] + ')')
        f.write("\n")
        for n in range(nimg):
            f.write('----For image: ' + imagefile[n] + '----')
            f.write('\n')
            chans_ = chans[n]
            for i in range(len(chans_)):
                f.write(str(chans_[i]) + ', ' + str(bmaj[n][i]) + ', ' + str(bmin[n][i]) + ', ' + str(bpa[n][i]))
                f.write("\n")
        f.close()
    return bmaj, bmin, bpa, beamunit, bpaunit


def imreg(vis=None, ephem=None, msinfo=None, imagefile=None, timerange=None, reftime=None, fitsfile=None, fitsdir='',beamfile=None,
          offsetfile=None, toTb=None, sclfactor=1.0, verbose=False, p_ang=False, overwrite=True, usephacenter=True,
          deletehistory=False, subregion=[], docompress=False):
    ''' 
    main routine to register CASA images
           Required Inputs:
               vis: STRING. CASA measurement set from which the image is derived
               imagefile: STRING or LIST. name of the input CASA image
               timerange: STRING or LIST. timerange used to generate the CASA image, must have the same length as the input images. 
                          Each element should be in CASA standard time format, e.g., '2012/03/03/12:00:00~2012/03/03/13:00:00'
           Optional Inputs:
               msinfo: DICTIONARY. CASA MS information, output from read_msinfo. If not provided, generate one from the supplied vis
               ephem: DICTIONARY. solar ephem, output from read_horizons. 
                      If not provided, query JPL Horizons based on time info of the vis (internet connection required)
               fitsfile: STRING or LIST. name of the output registered fits files
               reftime: STRING or LIST. Each element should be in CASA standard time format, e.g., '2012/03/03/12:00:00'
               offsetfile: optionally provide an offset with a series of solar x and y offsets with timestamps 
               toTb: Bool. Convert the default Jy/beam to brightness temperature?
               sclfactor: scale the image values up by its value (to compensate VLA 20 dB attenuator)
               verbose: Bool. Show more diagnostic info if True.
               usephacenter: Bool -- if True, correct for the RA and DEC in the ms file based on solar empheris.
                                     Otherwise assume the phasecenter is correctly pointed to the solar disk center
                                     (EOVSA case)
               subregion: Region selection. See 'help par.region' for details.
    Usage:
    >>> from suncasa.utils import helioimage2fits as hf
    >>> hf.imreg(vis='mydata.ms', imagefile='myimage.image', fitsfile='myimage.fits',
                 timerange='2017/08/21/20:21:10~2017/08/21/20:21:18')
    The output fits file is 'myimage.fits'

    History:
    BC (sometime in 2014): function was first wrote, followed by a number of edits by BC and SY
    BC (2019-07-16): Added checks for stokes parameter. Verified that for converting from Jy/beam to brightness temperature,
                     the convention of 2*k_b*T should always be used. I.e., for unpolarized source, stokes I, RR, LL, XX, YY, 
                     etc. in the output CASA images from (t)clean should all have same values of radio intensity 
                     (in Jy/beam) and brightness temperature (in K).

    '''
    #ia = iatool()

    if deletehistory:
        ms_clearhistory(vis)

    if not imagefile:
        raise ValueError('Please specify input image')
    if not timerange:
        raise ValueError('Please specify timerange of the input image')
    if type(imagefile) == str:
        imagefile = [imagefile]
    if type(timerange) == str:
        timerange = [timerange]
    if not fitsfile:
        fitsfile = [os.path.join(fitsdir,os.path.basename(img)) + '.fits' for img in imagefile]
    if type(fitsfile) == str:
        fitsfile = [os.path.join(fitsdir,fitsfile)]
    nimg = len(imagefile)
    if len(timerange) != nimg:
        raise ValueError('Number of input images does not equal to number of timeranges!')
    if len(fitsfile) != nimg:
        raise ValueError('Number of input images does not equal to number of output fits files!')
    nimg = len(imagefile)
    if verbose:
        print(str(nimg) + ' images to process...')

    if reftime:  # use as reference time to find solar disk RA and DEC to register the image, but not the actual timerange associated with the image
        if type(reftime) == str:
            reftime = [reftime] * nimg
        if len(reftime) != nimg:
            raise ValueError('Number of reference times does not match that of input images!')
        helio = ephem_to_helio(vis, ephem=ephem, msinfo=msinfo, reftime=reftime, usephacenter=usephacenter)
    else:
        # use the supplied timerange to register the image
        helio = ephem_to_helio(vis, ephem=ephem, msinfo=msinfo, reftime=timerange, usephacenter=usephacenter)

    if toTb:
        (bmajs, bmins, bpas, beamunits, bpaunits) = getbeam(imagefile=imagefile, beamfile=beamfile)

    for n, img in enumerate(imagefile):
        if verbose:
            print('processing image #' + str(n) + ' ' + img)
        fitsf = fitsfile[n]
        timeran = timerange[n]
        # obtain duration of the image as FITS header exptime
        try:
            [tbg0, tend0] = timeran.split('~')
            tbg_d = qa.getvalue(qa.convert(qa.totime(tbg0), 'd'))[0]
            tend_d = qa.getvalue(qa.convert(qa.totime(tend0), 'd'))[0]
            tdur_s = (tend_d - tbg_d) * 3600. * 24.
            dateobs = qa.time(qa.quantity(tbg_d, 'd'), form='fits', prec=10)[0]
        except:
            print('Error in converting the input timerange: ' + str(timeran) + '. Proceeding to the next image...')
            continue

        hel = helio[n]
        if not os.path.exists(img):
            warnings.warn('{} does not existed!'.format(img))
        else:
            if os.path.exists(fitsf) and not overwrite:
                raise ValueError('Specified fits file already exists and overwrite is set to False. Aborting...')
            else:
                p0 = hel['p0']
                tb.open(img + '/logtable', nomodify=False)
                nobs = tb.nrows()
                tb.removerows([i + 1 for i in range(nobs - 1)])
                tb.close()
                ia.open(img)
                imr = ia.rotate(pa=str(-p0) + 'deg')
                if subregion is not []:
                    imr = imr.subimage(region=subregion)
                imr.tofits(fitsf, history=False, overwrite=overwrite)
                imr.close()
                imsum = ia.summary()
                ia.close()
                ia.done()

            # construct the standard fits header
            # RA and DEC of the reference pixel crpix1 and crpix2
            (imra, imdec) = (imsum['refval'][0], imsum['refval'][1])
            # find out the difference of the image center to the CASA phase center
            # RA and DEC difference in arcseconds
            ddec = degrees((imdec - hel['dec_fld'])) * 3600.
            dra = degrees((imra - hel['ra_fld']) * cos(hel['dec_fld'])) * 3600.
            # Convert into image heliocentric offsets
            prad = -radians(hel['p0'])
            dx = (-dra) * cos(prad) - ddec * sin(prad)
            dy = (-dra) * sin(prad) + ddec * cos(prad)
            if offsetfile:
                try:
                    offset = np.load(offsetfile)
                except:
                    raise ValueError('The specified offsetfile does not exist!')
                reftimes_d = offset['reftimes_d']
                xoffs = offset['xoffs']
                yoffs = offset['yoffs']
                timg_d = hel['reftime']
                ind = bisect.bisect_left(reftimes_d, timg_d)
                xoff = xoffs[ind - 1]
                yoff = yoffs[ind - 1]
            else:
                xoff = hel['refx']
                yoff = hel['refy']
            if verbose:
                print('offset of image phase center to visibility phase center (arcsec): dx={0:.2f}, dy={1:.2f}'.format(
                    dx, dy))
                print('offset of visibility phase center to solar disk center (arcsec): dx={0:.2f}, dy={1:.2f}'.format(
                    xoff, yoff))
            (crval1, crval2) = (xoff + dx, yoff + dy)
            # update the fits header to heliocentric coordinates

            hdu = pyfits.open(fitsf, mode='update')
            hdu[0].verify('fix')
            header = hdu[0].header
            dshape = hdu[0].data.shape
            ndim = hdu[0].data.ndim
            (cdelt1, cdelt2) = (
                -header['cdelt1'] * 3600., header['cdelt2'] * 3600.)  # Original CDELT1, 2 are for RA and DEC in degrees
            header['cdelt1'] = cdelt1
            header['cdelt2'] = cdelt2
            header['cunit1'] = 'arcsec'
            header['cunit2'] = 'arcsec'
            header['crval1'] = crval1
            header['crval2'] = crval2
            header['ctype1'] = 'HPLN-TAN'
            header['ctype2'] = 'HPLT-TAN'
            header['date-obs'] = dateobs  # begin time of the image
            if not p_ang:
                hel['p0'] = 0
            try:
                # this works for pyfits version of CASA 4.7.0 but not CASA 4.6.0
                if tdur_s:
                    header.set('exptime', tdur_s)
                else:
                    header.set('exptime', 1.)
                header.set('p_angle', hel['p0'])
                header.set('dsun_obs', coordinates.sun.earth_distance(Time(dateobs)).to(u.meter).value)
                header.set('rsun_obs', coordinates.sun.angular_radius(Time(dateobs)).value)
                header.set('rsun_ref', sun.constants.radius.value)
                header.set('hgln_obs', 0.)
                header.set('hglt_obs', coordinates.sun.B0(Time(dateobs)).value)
            except:
                # this works for astropy.io.fits
                if tdur_s:
                    header.append(('exptime', tdur_s))
                else:
                    header.append(('exptime', 1.))
                header.append(('p_angle', hel['p0']))
                header.append(('dsun_obs', coordinates.sun.earth_distance(Time(dateobs)).to(u.meter).value))
                header.append(('rsun_obs', coordinates.sun.angular_radius(Time(dateobs)).value))
                header.append(('rsun_ref', sun.constants.radius.value))
                header.append(('hgln_obs', 0.))
                header.append(('hglt_obs', coordinates.sun.B0(Time(dateobs))[1].value))

            # check if stokes parameter exist
            exist_stokes = False
            stokes_mapper = {1:'I', 2:'Q', 3:'U', 4:'V', -1:'RR', -2:'LL',
                             -3:'RL', -4:'LR', -5:'XX', -6:'YY', -7:'XY', -8:'YX'}
            if 'CRVAL3' in header.keys():
                if header['CTYPE3'] == 'STOKES':
                    stokenum = header['CRVAL3']
                    exist_stokes = True
            if 'CRVAL4' in header.keys():
                if header['CTYPE4'] == 'STOKES':
                    stokenum = header['CRVAL4']
                    exist_stokes = True
            if exist_stokes:
                stokesstr = stokes_mapper[stokenum]
                if verbose:
                    print('This image is in Stokes ' + stokesstr)
            else:
                print('STOKES Information does not seem to exist! Assuming Stokes I')
                stokenum = 1

            # intensity units to brightness temperature
            if toTb:
                # get restoring beam info
                bmaj = bmajs[n]
                bmin = bmins[n]
                beamunit = beamunits[n]
                data = hdu[0].data  # remember the data order is reversed due to the FITS convension
                keys = header.keys()
                values = header.values()
                # which axis is frequency?
                faxis = keys[values.index('FREQ')][-1]
                faxis_ind = ndim - int(faxis)
                # find out the polarization of this image
                k_b = qa.constants('k')['value']
                c_l = qa.constants('c')['value']
                # Always use 2*kb for all polarizations
                const = 2. * k_b / c_l ** 2
                if header['BUNIT'].lower() == 'jy/beam':
                    header['BUNIT'] = 'K'
                    header['BTYPE'] = 'Brightness Temperature'
                    for i in range(dshape[faxis_ind]):
                        nu = header['CRVAL' + faxis] + header['CDELT' + faxis] * (i + 1 - header['CRPIX' + faxis])
                        if header['CUNIT' + faxis] == 'KHz':
                            nu *= 1e3
                        if header['CUNIT' + faxis] == 'MHz':
                            nu *= 1e6
                        if header['CUNIT' + faxis] == 'GHz':
                            nu *= 1e9
                        if len(bmaj) > 1:  # multiple (per-plane) beams
                            bmajtmp = bmaj[i]
                            bmintmp = bmin[i]
                        else:  # one single beam
                            bmajtmp = bmaj[0]
                            bmintmp = bmin[0]
                        if beamunit == 'arcsec':
                            bmaj0 = np.radians(bmajtmp / 3600.)
                            bmin0 = np.radians(bmintmp / 3600.)
                        if beamunit == 'arcmin':
                            bmaj0 = np.radians(bmajtmp / 60.)
                            bmin0 = np.radians(bmintmp / 60.)
                        if beamunit == 'deg':
                            bmaj0 = np.radians(bmajtmp)
                            bmin0 = np.radians(bmintmp)
                        if beamunit == 'rad':
                            bmaj0 = bmajtmp
                            bmin0 = bmintmp
                        beam_area = bmaj0 * bmin0 * np.pi / (4. * log(2.))
                        factor = const * nu ** 2  # SI unit
                        jy_to_si = 1e-26
                        # print(nu/1e9, beam_area, factor)
                        factor2 = sclfactor
                        # if sclfactor:
                        #     factor2 = 100.
                        if faxis == '3':
                            data[:, i, :, :] *= jy_to_si / beam_area / factor * factor2
                        if faxis == '4':
                            data[i, :, :, :] *= jy_to_si / beam_area / factor * factor2

            header = fu.headerfix(header)
            hdu.flush()
            hdu.close()

            if ndim - np.count_nonzero(np.array(dshape) == 1) > 3:
                docompress = False
                '''
                    Caveat: only 1D, 2D, or 3D images are currently supported by
                    the astropy fits compression. If a n-dimensional image data array
                    does not have at least n-3 single-dimensional entries,
                    force docompress to be False
                '''

                print('warning: The fits data contains more than 3 non squeezable dimensions. Skipping fits compression..')
            if docompress:
                fitsftmp = fitsf + ".tmp.fits"
                os.system("mv {} {}".format(fitsf, fitsftmp))
                hdu = pyfits.open(fitsftmp)
                hdu[0].verify('fix')
                header = hdu[0].header
                data = hdu[0].data
                fu.write_compressed_image_fits(fitsf, data, header, compression_type='RICE_1',
                                             quantize_level=4.0)
                os.system("rm -rf {}".format(fitsftmp))
    if deletehistory:
        ms_restorehistory(vis)
    return fitsfile


def calc_phasecenter_from_solxy(vis, timerange='', xycen=None, usemsphacenter=True):
    '''
    return the phase center in RA and DEC of a given solar coordinates

    :param vis: input measurement sets file
    :param timerange: can be a string or astropy.time.core.Time object, or a 2-element list of string or Time object
    :param xycen:  solar x-pos and y-pos in arcsec
    :param usemsphacenter:
    :return:
    phasecenter
    midtim: mid time of the given timerange
    '''
    tb.open(vis + '/POINTING')
    tst = Time(tb.getcell('TIME_ORIGIN', 0) / 24. / 3600., format='mjd')
    ted = Time(tb.getcell('TIME_ORIGIN', tb.nrows() - 1) / 24. / 3600., format='mjd')
    tb.close()
    datstr = tst.iso[:10]

    if isinstance(timerange, Time):
        try:
            (sttim, edtim) = timerange
        except:
            sttim = timerange
            edtim = sttim
    else:
        if timerange == '':
            sttim = tst
            edtim = ted
        else:
            try:
                (tstart, tend) = timerange.split('~')
                if tstart[2] == ':':
                    sttim = Time(datstr + 'T' + tstart)
                    edtim = Time(datstr + 'T' + tend)
                    # timerange = '{0}/{1}~{0}/{2}'.format(datstr.replace('-', '/'), tstart, tend)
                else:
                    sttim = Time(qa.quantity(tstart, 'd')['value'], format='mjd')
                    edtim = Time(qa.quantity(tend, 'd')['value'], format='mjd')
            except:
                try:
                    if timerange[2] == ':':
                        sttim = Time(datstr + 'T' + timerange)
                        edtim = sttim
                    else:
                        sttim = Time(qa.quantity(timerange, 'd')['value'], format='mjd')
                        edtim = sttim
                except ValueError:
                    print("keyword 'timerange' in wrong format")

    ms.open(vis)
    metadata = ms.metadata()
    observatory = metadata.observatorynames()[0]
    ms.close()

    midtim_mjd = (sttim.mjd + edtim.mjd) / 2.
    midtim = Time(midtim_mjd, format='mjd')
    eph = read_horizons(t0=midtim)
    if observatory == 'EOVSA' or (not usemsphacenter):
        print('This is EOVSA data')
        # use RA and DEC from FIELD ID 0
        tb.open(vis + '/FIELD')
        phadir = tb.getcol('PHASE_DIR').flatten()
        tb.close()
        ra0 = phadir[0]
        dec0 = phadir[1]
    else:
        ra0 = eph['ra'][0]
        dec0 = eph['dec'][0]

    if not xycen:
        # use solar disk center as default
        phasecenter = 'J2000 ' + str(ra0) + 'rad ' + str(dec0) + 'rad'
    else:
        x0 = np.radians(xycen[0] / 3600.)
        y0 = np.radians(xycen[1] / 3600.)
        p0 = np.radians(eph['p0'][0])  # p angle in radians
        raoff = -((x0) * np.cos(p0) - y0 * np.sin(p0)) / np.cos(eph['dec'][0])
        decoff = (x0) * np.sin(p0) + y0 * np.cos(p0)
        newra = ra0 + raoff
        newdec = dec0 + decoff
        phasecenter = 'J2000 ' + str(newra) + 'rad ' + str(newdec) + 'rad'
    return phasecenter, midtim



## only works in CASA 5
from astropy.time import Time
import numpy as np
import os

def make_ephem(vis, ephemfile=None):
    import urllib2, ssl
    from taskinit import tb
    quantities = ['1', '14', '15', '17', '19', '20', '24', '32']
    quantities = ','.join(quantities)
    tb.open(vis)
    btime = Time(tb.getcell('TIME', 0) / 24. / 3600., format='mjd')
    etime = Time(tb.getcell('TIME', tb.nrows() - 1) / 24. / 3600., format='mjd')
    tb.close()
    print("Beginning time of this scan " + btime.iso)
    print("End time of this scan " + etime.iso)

    btime = Time((btime.mjd - 1.0/60./24.), format='mjd')
    etime = Time((etime.mjd + 1.0/60./24.), format='mjd')
    startdate = btime.iso.replace(' ', ',')[:-7]
    enddate = etime.iso.replace(' ', ',')[:-7]
    cmd = ["COMMAND= '10'", "CENTER= '-5@399'", "MAKE_EPHEM= 'YES'", "TABLE_TYPE= 'OBSERVER'",
           "START_TIME= '%s'" % startdate, "STOP_TIME= '%s'" % enddate, "STEP_SIZE= '1m'", "CAL_FORMAT= 'CAL'",
           "TIME_DIGITS= 'MINUTES'", "ANG_FORMAT= 'DEG'", "OUT_UNITS= 'KM-S'", "RANGE_UNITS= 'AU'",
           "APPARENT= 'AIRLESS'", "SOLAR_ELONG= '0,180'", "SUPPRESS_RANGE_RATE= 'NO'", "SKIP_DAYLT= 'NO'",
           "EXTRA_PREC= 'NO'", "R_T_S_ONLY= 'NO'", "REF_SYSTEM= 'J2000'", "CSV_FORMAT= 'YES'", "OBJ_DATA= 'YES'",
           "TIME_DIGITS ='MIN'", "QUANTITIES= '{}'".format(quantities)]
    cmdstr = "http://ssd.jpl.nasa.gov/horizons_batch.cgi?batch=l&" + '&'.join(cmd)
    try:
        context = ssl._create_unverified_context()
        f = urllib2.urlopen(cmdstr, context=context)
    except:
        f = urllib2.urlopen(cmdstr)
    lines = f.readlines()
    f.close()
    istart = 0
    for i, l in enumerate(lines):
        if l[0:5] == '$$SOE':  # start recording
            istart = i + 1
        if l[0:5] == '$$EOE':  # end recording
            iend = i

    if not ephemfile:
        ephemfile = 'sun-ephem-geo.txt'
    with open(ephemfile, 'w') as fb:
        for i, l in enumerate(lines):
            if i == istart - 3:
                fb.write(
                    ' Date__(UT)__HR:MN     R.A.___(J2000.0)___DEC. Ob-lon Ob-lat Sl-lon Sl-lat   NP.ang   NP.dist               r        rdot            delta      deldot    S-T-O')
            if i >= istart and i < iend:
                l_s = l.split(',')
                l_s.pop(1)
                l_s.pop(1)
                fb.write(' '.join(l_s))
            else:
                fb.write(l)
                # with open(ephemfile,'w') as fb:
                #     for i,l in enumerate(lines):
                #         fb.write(l.replace('*m','').replace('*t',''))


def make_ephem_tb(vis, output_casa_table_name=None):
    import recipes.ephemerides.JPLephem_reader2 as jplreader
    import os.path
    ephemfile = 'sun-ephem.txt'
    make_ephem(vis, ephemfile=ephemfile)
    if os.path.isfile(ephemfile):
        output_dictionary = jplreader.readJPLephem(ephemfile)
        time_range_string = str(int(output_dictionary['earliest']['m0']['value'])) + '-' + str(
            int(output_dictionary['latest']['m0']['value'])) + 'dUTC'
        if not output_casa_table_name:
            output_casa_table_name = '_'.join([output_dictionary['NAME'], time_range_string, 'J2000.tab'])
        jplreader.ephem_dict_to_table(output_dictionary, output_casa_table_name)
        print('Ephemeris table written to {}'.format(output_casa_table_name))
    else:
        print("sun-ephem-geo.txt not found!")

def ephem2evla(ephemfile):
    import recipes.ephemerides.JPLephem_reader2 as jplreader
    if os.path.isfile(ephemfile):
        output_dictionary = jplreader.readJPLephem(ephemfile)
        print('RA DEC  table written to {}'.format(output_casa_table_name))
        np.deg2rad(output_dictionary['data']['DEC']['data']['value'])
        np.deg2rad(output_dictionary['data']['RA']['data']['value'])
        output_dictionary['data']['Rho']['data']['value']
    else:
        print("sun-ephem-geo.txt not found!")


from astropy.time import Time
import numpy as np
"""
Updated version to support the current JPL-Horizons query results as of 2013 June.
(when S-T-O is requested the current query result returns phi, PAB-LON, PAB-LAT along
with S-T-O value).  

casapy functions for converting ASCII ephemerides from JPL-Horizons into
CASA tables and installing them where casapy can find them.
                    
jplfiles_to_repository() puts it all together, so it is most likely the
function you want.

There are various utilities like convert_radec, datestr*, get_num_from_str,
mean_radius*, and construct_tablepath defined in here as well.
"""

from glob import glob
import os
import re
import scipy.special
import time                  # We can always use more time.

from taskinit import gentools, qa, casalog
me = gentools(['me'])[0]
from dict_to_table import dict_to_table

# Possible columns, as announced by their column titles.
# The data is scooped up by 'pat'.  Either use ONE group named by the column
# key, or mark it as unwanted.  '-' is not valid in group names.
# Leading and trailing whitespace will be taken care of later.
# Sample lines:
#  Date__(UT)__HR:MN     R.A.___(ICRF/J2000.0)___DEC Ob-lon Ob-lat Sl-lon Sl-lat   NP.ang   NP.dist               r        rdot            delta      deldot    S-T-O   L_s
#  2010-May-01 00:00     09 01 43.1966 +19 04 28.673 286.52  18.22 246.99  25.34 358.6230      3.44  1.661167637023  -0.5303431 1.28664311447968  15.7195833  37.3033   84.50
#
# some mod to label names and comments so that they corresponds to
# JPL-Horizons naming comvension
cols = {
    'MJD': {'header': r'Date__\(UT\)__HR:MN(:)?(\w+)?(\.)?(\w+)?',
            'comment': 'date',
            'pat':     r'(?P<MJD>\d+-\w+-\d+ \d+:\d+)'},
    'RA': {'header': r'R.A._+\([^)]+',
           'comment': 'Right Ascension',
           'pat':    r'(?P<RA>(\d+ \d+ )?\d+\.\d+)'}, # require a . for safety
    'DEC': {'header': r'\)_+DEC.',
            'comment': 'Declination',
            'pat':    r'(?P<DEC>([-+]?\d+ \d+ )?[-+]?\d+\.\d+)'},
    'NP_ang': {'header': r'NP\.ang',
               'comment': 'North-Pole pos. angle',
               'pat': r'(?P<NP_ang>[0-9.]+|n\.a\.)',
               'unit': 'deg'},
    'NP_dist': {'header': r'NP\.dist',
                'comment': 'North-Pole ang. distance',
               'pat': r'(?P<NP_dist>[-+0-9.]+|n\.a\.)',
                'unit': 'arcsec'},
    'illu': {'header': r'Illu%',
             #'comment': 'Illumination',
             'comment': 'Illuminated fraction',
             'pat':    r'(?P<illu>[0-9.]+)',
             'unit': r'%'},
    # put back to original heading...
    'DiskLong': {'header': r'Ob-lon',
    #'Obs_Long': {'header': r'Ob-lon',
               'comment': 'Sub-observer longitude',
    #           'pat':    r'(?P<Obs_Long>[0-9.]+|n\.a\.)',
               'pat':    r'(?P<DiskLong>[0-9.]+|n\.a\.)',
               'unit': 'deg'},
    'DiskLat': {'header': r'Ob-lat',
    #'Obs_Lat': {'header': r'Ob-lat',
               'comment': 'Sub-observer latitude',
    #           'pat':    r'(?P<Obs_Lat>[-+0-9.]+|n\.a\.)',
               'pat':    r'(?P<DiskLat>[-+0-9.]+|n\.a\.)',
               'unit': 'deg'},
    'Sl_lon': {'header': r'Sl-lon',
               'comment': 'Sub-Solar longitude',
               'pat':    r'(?P<Sl_lon>[0-9.]+|n\.a\.)',
               'unit': 'deg'},
    'Sl_lat': {'header': r'Sl-lat',
               'comment': 'Sub-Solar longitude',
               'pat':    r'(?P<Sl_lat>[-+0-9.]+|n\.a\.)',
               'unit': 'deg'},

    # These are J2000 whether or not ra and dec are apparent directions.
    'NP_RA': {'header': r'N\.Pole-RA',
              'comment': 'North Pole right ascension',
              'pat':    r'(?P<NP_RA>(\d+ \d+ )?\d+\.\d+)'}, # require a . for safety
    'NP_DEC': {'header': r'N\.Pole-DC',
               'comment': 'North Pole declination',
               'pat':    r'(?P<NP_DEC>([-+]?\d+ \d+ )?[-+]?\d+\.\d+)'},

    'r': {'header': 'r',
          'comment': 'heliocentric distance',
          'unit':    'AU',
          'pat':     r'(?P<r>[0-9.]+)'},
    'rdot': {'header': 'rdot',
             #'comment': 'heliocentric velocity',
             'comment': 'heliocentric distance rate',
             'unit': 'km/s',
             'pat': r'(?P<rdot>[-+0-9.]+)'},
    #         'unwanted': True},
    'Rho': {'header': 'delta',
            'comment': 'geocentric distance',
            'unit':    'AU',
            'pat':     r'(?P<Rho>[0-9.]+)'},
    'RadVel': {'header': 'deldot',
               #'comment': 'Radial velocity relative to the observer',
               'comment': 'geocentric distance rate',
               'pat': r'(?P<RadVel>[-+0-9.]+)',
               'unit': 'km/s'},
    'phang': {'header':  'S-T-O',
              'comment': 'phase angle',
              'unit':    'deg',
              'pat':     r'(?P<phang>[0-9.]+)'},
    # added columns to match current(as of June 2013) query result
    'phi': {'header':  'phi',
	    'comment': 'phase angle',
	    'unit':    'deg',
            'pat':     r'(?P<phi>[0-9.]+)',
            'unwanted': True},
    'PAB_LON': {'header':  'PAB-LON',
	    'comment': 'ecliptic longitude',
	    'unit':    'deg',
            'pat':     r'(?P<PAB_LON>[0-9.]+)',
            'unwanted': True},
    'PAB_LAT': {'header':  'PAB-LAT',
	    'comment': 'ecliptic latitude',
	    'unit':    'deg',
            'pat':     r'(?P<PAB_LAT>[-+0-9.]+)',
            'unwanted': True},
    # -----------------------------------------------------------
    'ang_sep': {'header': 'ang-sep/v',
                'comment': 'Angular separation from primary',
                'pat': r'(?P<ang_sep>[0-9.]+/.)'},  # arcsec, "visibility code".
                                                    # t: transiting primary
                                                    # O: occulted by primary
                                                    # p: partial umbral eclipse
                                                    # P: occulted partial umbral eclipse
                                                    # u: total umbral eclipse
                                                    # U: occulted total umbral eclipse
                                                    # *: none of the above
                                                    # -: target is primary
    'L_s': {'header': 'L_s',  # 08/2010: JPL does not supply this and
            'unit': 'deg',    # says they cannot.  Ask Bryan Butler.
            'comment': 'Season angle',
            'pat': r'(?P<L_s>[-+0-9.]+)'}

    }

def readJPLephem(fmfile,version=''):
    """
    Reads a JPL Horizons text file (see
    http://ssd.jpl.nasa.gov/horizons.cgi#top ) for a solar system object and
    returns various quantities in a dictionary.  The dict will be blank ({}) if
    there is a failure.
    """
    retdict = {}
    casalog.origin('readJPLephem')

    # Try opening fmfile now, because otherwise there's no point continuing.
    try:
        ephem = open(fmfile, 'rb')
        print("opened the file=",fmfile)
        lines=ephem.readlines()
        # skip this, handle by rstrip later
        #crCount=0
        #newlines=''
        #newln=''
        #for ln in lines:
        #  n = ln.count('\r')
        #  if n > 0:
        #    newln=ln.replace('\r','\n')
        #    crCount+=n
        #    newlines += newln
        #if crCount > 0:
        #  print("The input file appears to contains the carriage return code, \'^M\', will replace it with \'\\n\'...")
        #  raw_input('pause0')
        #  ephem.close()
        #  ephem = open('temp_ephem_data.txt','w+')
        #  ephem.write(newlines)
        ephem.seek(0)
    except IOError:
        casalog.post("Could not open ephemeris file " + fmfile,
                     priority="SEVERE")
        return {}

    # reset to default search pattern for MJD
    cols['MJD']['pat']=r'(?P<MJD>\d+-\w+-\d+ \d+:\d+)'
    # Setup the regexps.

    # Headers (one time only things)

    # Dictionary of quantity label: regexp pattern pairs that will be searched
    # for once.  The matching quantity will go in retdict[label].  Only a
    # single quantity (group) will be retrieved per line.
    headers = {
        'NAME': {'pat': r'^[>\s]*Target body name:\s+\d*\s*(\w+)'},   # object name, w.o. number
        'ephtype': {'pat': r'\?s_type=1#top>\]\s*:\s+\*(\w+)'}, # e.g. OBSERVER
        'obsloc': {'pat': r'^[>\s]*Center-site name:\s+(\w+)'},        # e.g. GEOCENTRIC
        # Catch either an explicit mean radius or a solitary target radius.
        'meanrad': {'pat': r'(?:Mean radius \(km\)\s*=|^Target radii\s*:)\s*([0-9.]+)(?:\s*km)?\s*$',
                    'unit': 'km'},
        # Triaxial target radii
        #'radii': {'pat': r'Target radii\s*:\s*([0-9.]+\s*x\s*[0-9.]+\s*x\s*[0-9.]+)\s*km.*Equator, meridian, pole',
        'radii': {'pat': r'Target radii\s*:\s*([0-9.]+\s*x\s*[0-9.]+\s*x\s*[0-9.]+)\s*km.*Equator, meridian, pole|Target radii\s*:\s*([0-9.]+)\s*km\s*',
                  'unit': 'km'},
        'T_mean': {'pat': r'Mean Temperature \(K\)\s*=\s*([0-9.]+)',
                   'unit': 'K'},

         # Figure out the units later.
        'rot_per': {'pat': r'(?i)(?<!Inferred )\b(rot(ation(al)?|\.)?\s*per.*=\s*([-0-9.]+\s*[dhr]*|Synchronous))'},
        'orb_per': {'pat': r'Orbital period((, days)?\s*=\s*[-0-9.]+\s*[dhr](\s*\(?R\)?)?)'},

        # MeasComet does not read units for these! E-lon(deg),  Lat(deg),     Alt(km)
        'GeoLong': {'pat': r'^[>\s]*Center geodetic\s*: ([-+0-9.]+,\s*[-+0-9.]+,\s*[-+0-9.]+)'},
        'dMJD':    {'pat': r'^[>\s]*Step-size\s*:\s*(.+)'},

        #                     request method v  wday mth   mday  hh  mm  ss   yyyy
        'VS_CREATE': {'pat': r'^[>\s]*Ephemeris / \w+ \w+ (\w+\s+\d+\s+\d+:\d+:\d+\s+\d+)'}
        }
    for hk in headers:
        headers[hk]['pat'] = re.compile(headers[hk]['pat'])

    # Data ("the rows of the table")

    # need date, r (heliocentric distance), delta (geocentric distance), and phang (phase angle).
    # (Could use the "dot" time derivatives for Doppler shifting, but it's
    # likely unnecessary.)
    #datapat = r'^[>\s]*'
    datapat = r'^\s*'

    stoppat = r'[>\s]*\$\$EOE$'  # Signifies the end of data.

    # Read fmfile into retdict.
    num_cols = 0
    in_data = False
    comp_mismatches = []
    print_datapat = False
    # define interpretation of invalid values ('n.a.')
    invalid=-999.
    for origline in ephem:
        line = origline.rstrip('\r\n')
        if in_data:
            if re.match(stoppat, line):
                break
            matchobj = re.search(datapat, line)
            if matchobj:
                #print("matchobj!")
                gdict = matchobj.groupdict()
                #print("gdict=",gdict)
                for col in gdict:
                    if gdict[col]=='n.a.':
                        gdict[col]=invalid
                #    print("cols.key=",cols.keys())

                    if not cols[col].get('unwanted'):
                        retdict['data'][col]['data'].append(gdict[col])
                if len(gdict) < num_cols:
                    print("Partially mismatching line:")
                    print(line)
                    print("Found:")
                    print(gdict)
                    print_datapat = True
                    raw_input("pause0")
            else:
                print_datapat = True
                # Chomp trailing whitespace.
                comp_mismatches.append(re.sub(r'\s*$', '', line))
        elif re.match(r'^[>\s]*' + cols['MJD']['header'] + r'\s+'
                      + cols['RA']['header'], line):
            # need to modify regex to search for the header name for MJD containing second digits
            if re.match(r'^[>\s]*Date__\(UT\)__HR:MN:SC.fff', line):
                cols['MJD']['pat'] = r'(?P<MJD>\d+-\w+-\d+ \d+:\d+:\d+.\d+)'
            # See what columns are present, and finish setting up datapat and
            # retdict.
            havecols = []
            # extract coordinate ref info

            m=re.match(r'(^[>\s]*)(\S+)(\s+)('+cols['RA']['header']+')', line)
            coordref=m.group(4).split('(')[-1]
            cols['RA']['comment']+='('+coordref+')'
            cols['DEC']['comment']+='('+coordref+')'
            #print("cols['RA']['comment']=",  cols['RA']['comment'])
            # Chomp trailing whitespace.
            myline = re.sub(r'\s*$', '', line)
            titleline = myline
            remaining_cols = cols.keys()
            found_col = True
            # This loop will terminate one way or another.
            while myline and remaining_cols and found_col:
                found_col = False
                #print("myline = '%s'" % myline)
                #print("remaining_cols =", ', '.join(remaining_cols))
                for col in remaining_cols:
                    if re.match(r'^[>\s]*' + cols[col]['header'], myline):
                        #print("Found", col)
                        havecols.append(col)
                        remaining_cols.remove(col)
                        myline = re.sub(r'^[>\s]*' + cols[col]['header'],
                                        '', myline)
                        found_col = True
                        break
            datapat += r'\s+'.join([cols[col]['pat'] for col in havecols])
            sdatapat = datapat
            casalog.post("Found columns: " + ', '.join(havecols))
            datapat = re.compile(datapat)
            retdict['data'] = {}
            for col in havecols:
                if not cols[col].get('unwanted'):
                    retdict['data'][col] = {'comment': cols[col]['comment'],
                                            'data':    []}
            num_cols = len(retdict['data'])
        #elif re.match(r'^\$\$SOE\s*$', line):  # Start of ephemeris
        elif re.match(r'^[>\s]*\$\$SOE\s*$', line):  # Start of ephemeris
            casalog.post("Starting to read data.", priority='INFO2')
            in_data = True
        else:
            #print("line =", line)
            #print("looking for",)
            for hk in headers:
                 #print("hk=",hk)

                 if not retdict.has_key(hk):
                    matchobj = re.search(headers[hk]['pat'], line)
                    if matchobj:
                        if hk=='radii':
                            mobjs=matchobj.groups()
                            for gp in mobjs:
                                if gp!=None:
                                    retdict[hk] = gp
                                    break
                            break
                        else:
                            retdict[hk] = matchobj.group(1) # 0 is the whole line
                            break
    ephem.close()
    # clean up the temp file if exists
    #if os.path.exists('temp_ephem_data.txt'):
    #  os.remove('temp_ephem_data.txt')

    # If there were errors, provide debugging info.
    if comp_mismatches:
        print("Completely mismatching lines:")
        #print("\n".join(comp_mismatches))
    if print_datapat:
        print("The apparent title line is:")
        print(titleline)
        print("datapat = r'%s'" % sdatapat)

    # Convert numerical strings into actual numbers.
    try:
        retdict['earliest'] = datestr_to_epoch(retdict['data']['MJD']['data'][0])
        retdict['latest'] = datestr_to_epoch(retdict['data']['MJD']['data'][-1])
    except Exception as e:
        print("Error!")
        if retdict.has_key('data'):
            if retdict['data'].has_key('MJD'):
                if retdict['data']['MJD'].has_key('data'):
                    #print("retdict['data']['MJD']['data'] =", retdict['data']['MJD']['data'])
                    print("retdict['data'] =", retdict['data'])
                else:
                    print("retdict['data']['MJD'] has no 'data' key.")
                    print("retdict['data']['MJD'].keys() =", retdict['data']['MJD'].keys())
            else:
                print("retdict['data'] has no 'MJD' key.")
                print("retdict['data'].keys() =", retdict['data'].keys())
        else:
            print("retdict has no 'data' key.")
        raise e

    for hk in headers:
        if retdict.has_key(hk):
            if headers[hk].has_key('unit'):
                if hk == 'radii':
                    radii = retdict[hk].split('x')
                    if len(radii)==1:
                        a = float(radii[0])
                        retdict[hk] = {'unit': headers[hk]['unit'], 'value': (a,a,a)}
                        retdict['meanrad'] = {'unit': headers[hk]['unit'],
                                          'value': a}
                    else:
                        a, b, c = [float(r) for r in radii]
                        retdict[hk] = {'unit': headers[hk]['unit'],
                                   'value': (a, b, c)}
                        retdict['meanrad'] = {'unit': headers[hk]['unit'],
                                          'value': mean_radius(a, b, c)}
                else:
                    try:
                        # meanrad might already have been converted.
                        if type(retdict[hk]) != dict:
                            retdict[hk] = {'unit': headers[hk]['unit'],
                                           'value': float(retdict[hk])}
                    except Exception as e:
                        print("Error converting header", hk, "to a Quantity.")
                        print("retdict[hk] =", retdict[hk])
                        raise e
            elif hk == 'GeoLong':
                long_lat_alt = retdict[hk].split(',')
                retdict['GeoLong'] = float(long_lat_alt[0])
                retdict['GeoLat']  = float(long_lat_alt[1])
                retdict['GeoDist'] = float(long_lat_alt[2])
            elif hk == 'dMJD':
                retdict[hk] = qa.convert(qa.totime(retdict[hk].replace('minutes', 'min')),
                                         'd')['value']
            elif hk == 'orb_per':
                unit = 'h'
                retrograde = False
                if 'd' in retdict[hk].lower():
                    unit = 'd'                 # Actually this is most common.
                if 'r' in retdict[hk].lower():
                    retrograde = True
                value = get_num_from_str(retdict[hk], 'orbital period')
                if value != False:
                    if retrograde and value > 0.0:
                        value = -value
                    retdict[hk] = {'unit': unit, 'value': value}
                else:
                    del retdict[hk]

    # The rotation period might depend on the orbital period ("Synchronous"),
    # so handle it after all the other headers have been done.
    if 'rot_per' in retdict:
        rpstr = retdict['rot_per']
        if 'ROTPER' in rpstr:                            # Asteroid
            retdict['rot_per'] = {'unit': 'h',         # Always seems to be for asteroids.
                                  'value': get_num_from_str(rpstr, 'rotation period')}
        elif 'Synchronous' in rpstr:
            retdict['rot_per'] = retdict['orb_per']
        else:  # Most likely a planet.
            match = re.search(r'(\d+)h\s*(\d+)m\s*([0-9.]+)s', rpstr)
            if match:
                hms = [float(match.group(i)) for i in range(1, 4)]
                retdict['rot_per'] = {'unit': 'h',
                                      'value': hms[0] + (hms[1] + hms[2] / 60.0) / 60.0}
            else:
                # DON'T include the optional r in hr!  qa.totime can't handle it.
                try:
                    match = re.search(r'([-0-9.]+)(?:\s*\+-[0-9.]+)?\s*([dh])', rpstr)
                    if match:
                        retdict['rot_per'] = {'unit': match.group(2),
                                              'value': float(match.group(1))}
                except:
                    print("Error parsing the rotation period from")
                    print(rpstr)

    if retdict['data'].has_key('ang_sep'):
        retdict['data']['obs_code'] = {'comment': 'Obscuration code'}
    for dk in retdict['data']:
        if dk == 'obs_code':
            continue
        if cols[dk].has_key('unit'):
            retdict['data'][dk]['data'] = {'unit': cols[dk]['unit'],
                      'value': scipy.array([float(s) for s in retdict['data'][dk]['data']])}
            if dk == 'RadVel':
                # Convert from km/s to AU/d.  Blame MeasComet, not me.
                retdict['data'][dk]['data']['unit'] = 'AU/d'
                kmps_to_AUpd = qa.convert('1km/s', 'AU/d')['value']
                retdict['data'][dk]['data']['value'] *= kmps_to_AUpd

        if re.match(r'.*(RA|DEC)$', dk):
            retdict['data'][dk] = convert_radec(retdict['data'][dk])
        elif dk == 'MJD':
            retdict['data']['MJD'] = datestrs_to_MJDs(retdict['data']['MJD'])
        elif dk == 'ang_sep':
            angseps = []
            obscodes = []
            for asoc in retdict['data'][dk]['data']:
                angsep, obscode = asoc.split('/')
                angseps.append(float(angsep))
                obscodes.append(obscode)
            retdict['data'][dk]['data'] = {'unit': 'arcseconds',
                                           'value': angseps}
            retdict['data']['obs_code']['data'] = obscodes

    if len(retdict.get('radii', {'value': []})['value']) == 3 \
           and retdict['data'].has_key('NP_RA') and retdict['data'].has_key('NP_DEC'):
        # Do a better mean radius estimate using the actual theta.
        retdict['meanrad']['value'] = mean_radius_with_known_theta(retdict)

    # To be eventually usable as a MeasComet table, a few more keywords are needed.
    retdict['VS_TYPE'] = 'Table of comet/planetary positions'
    if version=='':
        version='0003.0001'
    #retdict['VS_VERSION'] = '0003.0001'
    retdict['VS_VERSION'] = version
    if retdict.has_key('VS_CREATE'):
        dt = time.strptime(retdict['VS_CREATE'], "%b %d %H:%M:%S %Y")
    else:
        casalog.post("The ephemeris creation date was not found.  Using the current time.",
                     priority="WARN")
        dt = time.gmtime()
    retdict['VS_CREATE'] = time.strftime('%Y/%m/%d/%H:%M', dt)

    # VS_DATE is required by MeasComet, but it doesn't seem to be actually used.
    retdict['VS_DATE'] = time.strftime('%Y/%m/%d/%H:%M', time.gmtime())

    if retdict['data'].has_key('MJD'):
        #casalog.post("retdict.keys=%s" % retdict.keys())
        retdict['MJD0'] = retdict['data']['MJD']['value'][0] - retdict['dMJD']
    else:
        print("The table will not be usable with me.framecomet because it lacks MJD.")

    # adding posrefsys keyword
    if cols['RA']['comment'].count('J2000'):
        retdict['posrefsys']='ICRF/J2000.0'
    if cols['RA']['comment'].count('B1950'):
        retdict['posrefsys']='FK4/B1950.0'

    return retdict

def convert_radec(radec_col):
    """
    Returns a column of RAs or declinations as strings, radec_col, as a
    quantity column.  (Unfortunately MeasComet assumes the columns are
    Quantities instead of Measures, and uses GeoDist == 0.0 to toggle between
    APP and TOPO.)
    """
    angstrlist = radec_col['data']
    angq = {}
    nrows = len(angstrlist)

    if len(angstrlist[0].split()) > 1:
        # Prep angstrlist for qa.toangle()
        if radec_col['comment'][:len("declination")].lower() == 'declination':
            for i in range(nrows):
                dms = angstrlist[i].replace(' ', 'd', 1)
                angstrlist[i] = dms.replace(' ', 'm') + 's'
        else:                                                  # R.A.
            for i in range(nrows):
                angstrlist[i] = angstrlist[i].replace(' ', ':')

        # Do first conversion to get unit.
        try:
            firstang = qa.toangle(angstrlist[0])
        except Exception as e:
            print("Error: Could not convert", angstrlist[0], "to an angle.")
            raise e
        angq['unit'] = firstang['unit']
        angq['value'] = [firstang['value']]

        for angstr in angstrlist[1:]:
            angq['value'].append(qa.toangle(angstr)['value'])
    else:
        angq['unit'] = 'deg'                    # This is an assumption!
        angq['value'] = [float(a) for a in angstrlist]

    return {'comment': radec_col['comment'],
            'data': {'unit': angq['unit'],
                     'value': scipy.array(angq['value'])}}

def get_num_from_str(fstr, wanted="float"):
    """
    Like float(fstr) on steroids, in that it ignores things in fstr that aren't
    numbers.  Returns False on failure.

    wanted: an optional label for the type of number you wanted.
            Only used for distinguishing error messages.

    Example:
    >>> from JPLephem_reader import get_num_from_str
    >>> get_num_from_str('  Sidereal rot. period  =    58.6462 d  ')
    58.6462
    >>> get_num_from_str('Rotation period = 16.11+-0.01 hr', wanted='rotation period')
    16.109999999999999
    >>> get_num_from_str('Rotation period = Synchronous', wanted='rotation period')
    Could not convert "Rotation period = Synchronous" to a rotation period.
    False
    """
    match = re.search(r'([-+]?(\d+(\.\d*)?|\d*\.\d+)([eEdD][-+]?\d+)?)', fstr)
    if match:
        value = float(match.group(1))
    else:
        print("Could not convert \"%s\" to a %s." % (fstr, wanted))
        value = False
    return value

def mean_radius(a, b, c):
    """
    Return the average apparent mean radius of an ellipsoid with semiaxes
    a >= b >= c.
    "average" means average over time naively assuming the pole orientation
    is uniformly distributed over the whole sphere, and "apparent mean radius"
    means a radius that would give the same area as the apparent disk.
    """
    # This is an approximation, but it's not bad.
    # The exact equations for going from a, b, c, and the Euler angles to the
    # apparent ellipse are given in Drummond et al, Icarus, 1985a.
    # It's the integral over the spin phase that I have approximated, so the
    # approximation is exact for b == a, and appears to hold well for b << a.
    R = 0.5 * c**2 * (1.0 / b**2 + 1.0 / a**2)   # The magic ratio.
    if R < 0.95:
        sqrt1mR = scipy.sqrt(1.0 - R)
        # There is fake singularity (RlnR) at R = 0, but it is unlikely to
        # be a problem.
        try:
            Rterm = 0.5 * R * scipy.log((1.0 + sqrt1mR) / (1.0 - sqrt1mR)) / sqrt1mR
        except:
            Rterm = 0.0
    else:
        # Use a (rapidly converging) series expansion to avoid a fake
        # singularity at R = 1.
        Rterm = 1.0               # 0th order
        onemR = 1.0 - R
        onemRtothei = 1.0
        for i in range(1, 5):    # Start series at 1st order.
            onemRtothei *= onemR
            Rterm -= onemRtothei / (0.5 + 2.0 * i**2)
    avalfabeta = 0.5 * a * b * (1.0 + Rterm)
    return scipy.sqrt(avalfabeta)

def mean_radius_with_known_theta(retdict):
    """
    Return the average apparent mean radius of an ellipsoid with semiaxes
    a >= b >= c (= retdict['radii']['value']).
    "average" means average over a rotation period, and "apparent mean radius"
    means the radius of a circle with the same area as the apparent disk.
    """
    a = retdict['radii']['value'][0]
    b2 = retdict['radii']['value'][1]**2
    c2 = retdict['radii']['value'][2]**2
    onemboa2 = 1.0 - b2 / a**2
    units = {}
    values = {}
    for c in ['RA', 'DEC', 'NP_RA', 'NP_DEC']:
        units[c] = retdict['data'][c]['data']['unit']
        values[c] = retdict['data'][c]['data']['value']
    av = 0.0
    nrows = len(retdict['data']['RA']['data']['value'])
    for i in range(nrows):
        radec = me.direction('app', {'unit': units['RA'], 'value': values['RA'][i]},
                             {'unit': units['DEC'], 'value': values['DEC'][i]})
        np = me.direction('j2000', {'unit': units['NP_RA'], 'value': values['NP_RA'][i]},
                          {'unit': units['NP_DEC'], 'value': values['NP_DEC'][i]})
        szeta2 = scipy.sin(qa.convert(me.separation(radec, np), 'rad')['value'])**2
        csinz2 = c2 * szeta2
        bcosz2 = b2 * (1.0 - szeta2)
        bcz2pcsz2 = bcosz2 + csinz2
        m = csinz2 * onemboa2 / bcz2pcsz2
        av += (scipy.sqrt(bcz2pcsz2) * scipy.special.ellipe(m) - av) / (i + 1.0)
    return scipy.sqrt(2.0 * a * av / scipy.pi)

def datestr_to_epoch(datestr):
    """
    Given a UT date like "2010-May-01 00:00", returns an epoch measure.
    """
    return me.epoch(rf='UTC', v0=qa.totime(datestr))

def datestrs_to_MJDs(cdsdict):
    """
    All of the date strings must have the same reference frame (i.e. UT).
    """
    datestrlist = cdsdict['data']

    # Convert to FITS format, otherwise qa.totime() will silently drop the hours.
    datestrlist = [d.replace(' ', 'T') for d in datestrlist]

    timeq = {}
    # Do first conversion to get unit.
    firsttime = qa.totime(datestrlist[0])
    timeq['unit'] = firsttime['unit']
    timeq['value'] = [firsttime['value']]

    for datestr in datestrlist[1:]:
        timeq['value'].append(qa.totime(datestr)['value'])

    return {'unit': timeq['unit'],
            'value': scipy.array(timeq['value'])}

def construct_tablepath(fmdict, prefix=''):
    """
    Construct a suitable pathname for a CASA table made from fmdict,
    starting with prefix.  prefix can contain a /.

    If prefix is not given, it will be set to
    "ephem_JPL-Horizons_%s" % fmdict['NAME']
    """
    if not prefix:
        prefix = "ephem_JPL-Horizons_%s" % fmdict['NAME']
    return prefix + "_%.0f-%.0f%s%s.tab" % (fmdict['earliest']['m0']['value'],
                                            fmdict['latest']['m0']['value'],
                                            fmdict['latest']['m0']['unit'],
                                            fmdict['latest']['refer'])

def ephem_dict_to_table(fmdict, tablepath='', prefix=''):
    """
    Converts a dictionary from readJPLephem() to a CASA table, and attempts to
    save it to either to tablepath or a constructed directory name.
    Returns whether or not it was successful.

    If tablepath is blank and prefix is not given, the table will go to
    something like ephem_JPL-Horizons_NAME_EARLIEST-LATESTdUTC.tab.

    If tablepath is blank and prefix is given, the table will go to
    something like prefix_EARLIEST-LATESTdUTC.tab.  prefix can contain a /.
    """
    if not tablepath:
        tablepath = construct_tablepath(fmdict, prefix)
        print("Writing to", tablepath)

    # safe gaurd from zapping current directory by dict_to_table()
    if tablepath=='.' or tablepath=='./' or tablepath.isspace():
        raise Exception("Invalid tablepath: "+tablepath)
    retval = True
    # keepcolorder=T preserves column ordering in collist below
    #keepcolorder=False
    keepcolorder=True
    try:
        outdict = fmdict.copy() # Yes, I want a shallow copy.
        #kws = fmdict.keys()
        # reorder the keywords
        okws=['VS_CREATE','VS_DATE','VS_TYPE', 'VS_VERSION', 'NAME', 'MJD0', 'dMJD',
              'GeoDist', 'GeoLat', 'GeoLong', 'obsloc', 'posrefsys','earliest','latest',
              'radii','meanrad','orb_per','data']
        oldkws = fmdict.keys()
        kws=[]
        for ik in okws:
            if oldkws.count(ik):
              kws.append(ik)
              oldkws.remove(ik)
        kws+=oldkws
        kws.remove('data')
        collist = outdict['data'].keys()

        # For cosmetic reasons, encourage a certain order to the columns, i.e.
        # start with alphabetical order,
        collist.sort()
        # but put these ones first, in the listed order (ignore the reverse and
        # the pops) if they are present.
        #put_these_first = ['MJD', 'RA', 'DEC', 'Rho', 'RadVel', 'NP_RA', 'NP_DEC',
        #                   'DiskLong', 'DiskLat', 'sl_lon', 'sl_lat', 'r',
        #                   'ang_sep', 'obs_code']
        put_these_first = ['MJD', 'RA', 'DEC', 'Rho', 'RadVel', 'NP_ang', 'NP_dist',
        #                   'Obs_Long', 'Obs_Lat', 'Sl_lon', 'Sl_lat', 'r','rdot']
                           'DiskLong', 'DiskLat', 'Sl_lon', 'Sl_lat', 'r','rdot']
        # Like l.sort(), reverse() acts on its instance instead of returning a value.
        put_these_first.reverse()
        for c in put_these_first:
            if c in collist:
                collist.remove(c)
                collist.insert(0, c)

        clashing_cols = [c for c in collist if c in kws]
        if clashing_cols:
            raise ValueError('The input dictionary lists ' + ', '.join(clashing_cols) + ' as both keyword(s) and column(s)')

        # This promotes the keys in outdict['data'] up one level, and removes
        # 'data' as a key of outdict.
        outdict.update(outdict.pop('data'))

        # This is primarily because MeasComet insists on it, not because it
        # ever gets used.  Maybe subType should be changed to 'Asteroid',
        # 'Moon', or 'Planet', but I'm leaving it at 'Comet' for now.
        info = {'readme': 'Derived by JPLephem_reader.py from a JPL-Horizons ephemeris (http://ssd.jpl.nasa.gov/horizons.cgi#top)',
                'subType': 'Comet', 'type': 'IERS'}

        retval = dict_to_table(outdict, tablepath, kws, collist, info, keepcolorder)
    except Exception as e:
        print('Error', e, 'trying to export an ephemeris dict to', tablepath)
        retval = False

    return retval


def jplfiles_to_repository(objs, jpldir='.', jplext='.ephem',
                           log='null'):
    """
    For each Solar System object obj in the list objs,
        look for matching JPL-Horizons ASCII files with jplext in jpldir,
        read them into python dictionaries,
        write the dicts to CASA tables in $CASAROOT/data/ephemerides/JPL-Horizons/,
        and check that they can be read by me.framecomet().
    Returns the number of ephemerides processed + readable by me.framecomet.

    jpldir and jplext can be glob patterns.

    $CASAROOT is derived from $CASAPATH.

    Log messages will be directed to log for the duration of this function.
    Note that 'null' makes a NullLogSink, so it might be better than /dev/null.

    Example:
    import recipes.ephemerides.request as jplreq
    objs = jplreq.asteroids.keys() + jplreq.planets_and_moons.keys()
    jplfiles_to_repository(objs, os.getenv('CASAPATH').split()[0])
    """
    neph = 0
    casapath = os.getenv('CASAPATH')
    if not casapath:
        print("CASAPATH is not set.")
        return 0
    datadir = casapath.split()[0] + '/data/ephemerides/JPL-Horizons'
    if not os.path.isdir(datadir):
        try:
            os.mkdir(datadir)
            print("Created", datadir)
            print("You should probably svn add it.")
        except Exception as e:
            "Error", e, "creating", datadir
            return 0
    datadir += '/'

    #oldlog = casalog.logfile()
    # This is needed to stop WARN and above from printing to the console,
    # but it permanently severs the logger window.
    #casalog.setglobal(True)
    #casalog.setlogfile(log)

    if jpldir[-1] != '/':
        jpldir += '/'
    for sob in objs:
        capob = sob.capitalize()
        lob = sob.lower()
        jplfiles = glob(jpldir + lob + jplext) + glob(jpldir + capob + jplext)
        for jplfile in jplfiles:
            casalog.post('Reading ' + jplfile)
            fmdict = readJPLephem(jplfile)
            tabpath = construct_tablepath(fmdict, datadir + capob)
            ephem_dict_to_table(fmdict, tabpath)

            # Check if it is readable by me.framecomet.
            epoch = fmdict['earliest']
            epoch['m0']['value'] += 0.5 * (fmdict['latest']['m0']['value'] -
                                           epoch['m0']['value'])
            me.doframe(epoch)
            if me.framecomet(tabpath):
                neph += 1
            else:
                casalog.post(tabpath + " was not readable by me.framecomet.",
                             'WARN')

    #casalog.setlogfile(oldlog)

    return neph



def read_ephem(ephemfile):
    '''
    read pointing ephemeris file. The file should in the following format.

    $$SOE
    2016-Apr-09 16:00     01 14 08.5215  07 54 14.196 1.00178292520640   0.2051057
    $$EOE

    :param ephemfile:
    :return:
    '''
    from taskinit import qa, casalog
    from datetime import datetime
    # from analysisUtils import analysisUtils as au
    import re
    try:
        ephem = open(ephemfile, 'rb')
        print("opened the file=", ephemfile)
        lines = ephem.readlines()
        ephem.seek(0)
    except IOError:
        casalog.post("Could not open ephemeris file " + ephemfile,
                     priority="SEVERE")
        return {}
    stoppat = r'[>\s]*\$\$EOE$'  # Signifies the end of data.
    in_data = False
    tobjs = []
    ras = []
    decs = []
    for origline in ephem:
        line = origline.rstrip('\r\n')
        if in_data:
            if re.match(stoppat, line):
                break
            data = ' '.join(line.split())
            data = data.split(' ')
            tobjs.append(datetime.strptime(' '.join(data[:2]), "%Y-%b-%d %H:%M"))
            # ra, dec = au.radec2rad(':'.join(data[2:5]) + ' ' + ':'.join(data[5:8]))
            ra = qa.toangle('{x[0]}h{x[1]}m{x[2]}'.format(x=data[2:5]))['value']
            dec = qa.toangle('{x[0]}d{x[1]}m{x[2]}'.format(x=data[5:8]))['value']
            ras.append(ra)
            decs.append(dec)
        elif re.match(r'^[>\s]*\$\$SOE\s*$', line):  # Start of ephemeris
            casalog.post("Starting to read data.", priority='INFO2')
            in_data = True
        else:
            continue
    ephem.close()
    return Time(tobjs), np.array(ras), np.array(decs)


def make_ephem(vis, ephemfile=None):
    '''
    make ephemeris table from JPLhorizon:

    Cautioned that the tab created in CASA 5+ lacks many columns such as Rho, RadVel, etc., which are essential for fixplanets
    The issue comes from a update on jplreader after 2015. some keys are not write to the ephemeris table.
    This only works with CASA 4+
    :param vis:
    :param ephemfile:
    :return:
    '''
    import urllib2, ssl
    from taskinit import tb
    quantities = ['1', '14', '15', '17', '19', '20', '24', '32']
    quantities = ','.join(quantities)
    tb.open(vis + '/OBSERVATION')
    trs = {'BegTime': [], 'EndTime': []}
    for ll in range(tb.nrows()):
        tim0, tim1 = Time(tb.getcell('TIME_RANGE', ll) / 24 / 3600, format='mjd')
        trs['BegTime'].append(tim0)
        trs['EndTime'].append(tim1)
    tb.close()
    trs['BegTime'] = Time(trs['BegTime'])
    trs['EndTime'] = Time(trs['EndTime'])
    btime = np.min(trs['BegTime'])
    etime = np.max(trs['EndTime'])
    print("Beginning time of this scan " + btime.iso)
    print("End time of this scan " + etime.iso)

    btime = Time((btime.mjd - 1.0 / 60 / 24), format='mjd')
    etime = Time((etime.mjd + 1.0 / 60 / 24), format='mjd')
    startdate = btime.iso.replace(' ', ',')[:-7]
    enddate = etime.iso.replace(' ', ',')[:-7]

    ### Horizons Batch CGI Script
    ### The Horizons batch CGI script https://ssd.jpl.nasa.gov/horizons_batch.cgi is obsolete.
    ### lease instead use the Horizons API.
    # cmd = ["COMMAND= '10'", "CENTER= '-5@399'", "MAKE_EPHEM= 'YES'", "TABLE_TYPE= 'OBSERVER'",
    #        "START_TIME= '%s'" % startdate, "STOP_TIME= '%s'" % enddate, "STEP_SIZE= '1m'", "CAL_FORMAT= 'CAL'",
    #        "TIME_DIGITS= 'MINUTES'", "ANG_FORMAT= 'DEG'", "OUT_UNITS= 'KM-S'", "RANGE_UNITS= 'AU'",
    #        "APPARENT= 'AIRLESS'", "SOLAR_ELONG= '0,180'", "SUPPRESS_RANGE_RATE= 'NO'", "SKIP_DAYLT= 'NO'",
    #        "EXTRA_PREC= 'NO'", "R_T_S_ONLY= 'NO'", "REF_SYSTEM= 'J2000'", "CSV_FORMAT= 'YES'", "OBJ_DATA= 'YES'",
    #        "TIME_DIGITS ='MIN'", "QUANTITIES= '{}'".format(quantities)]
    # cmdstr = "http://ssd.jpl.nasa.gov/horizons_batch.cgi?batch=l&" + '&'.join(cmd)

    ### Horizons API
    cmd = ["COMMAND= '10'", "CENTER= '-5@399'", "MAKE_EPHEM= 'YES'", "TABLE_TYPE= 'OBSERVER'",
           "START_TIME= '%s'" % startdate, "STOP_TIME= '%s'" % enddate, "STEP_SIZE= '1m'", "CAL_FORMAT= 'CAL'",
           "TIME_DIGITS= 'MINUTES'", "ANG_FORMAT= 'DEG'", "OUT_UNITS= 'KM-S'", "RANGE_UNITS= 'AU'",
           "APPARENT= 'AIRLESS'", "SOLAR_ELONG= '0,180'", "SUPPRESS_RANGE_RATE= 'NO'", "SKIP_DAYLT= 'NO'",
           "EXTRA_PREC= 'NO'", "R_T_S_ONLY= 'NO'", "REF_SYSTEM= 'J2000'", "CSV_FORMAT= 'YES'", "OBJ_DATA= 'YES'",
           "QUANTITIES= '{}'".format(quantities)]
    cmdstr = "https://ssd.jpl.nasa.gov/api/horizons.api?format=text&" + '&'.join(cmd)


    cmdstr = cmdstr.replace("'", "%27")
    cmdstr = cmdstr.replace(" ", "%20")
    try:
        context = ssl._create_unverified_context()
        f = urllib2.urlopen(cmdstr, context=context)
    except:
        f = urllib2.urlopen(cmdstr)
    lines = f.readlines()
    f.close()
    istart = 0
    for i, l in enumerate(lines):
        if l[0:5] == '$$SOE':  # start recording
            istart = i + 1
        if l[0:5] == '$$EOE':  # end recording
            iend = i

    if not ephemfile:
        ephemfile = 'sun-ephem-geo.txt'
    with open(ephemfile, 'w') as fb:
        for i, l in enumerate(lines):
            if i == istart - 3:
                fb.write(
                    ' Date__(UT)__HR:MN     R.A.___(J2000.0)___DEC. Ob-lon Ob-lat Sl-lon Sl-lat   NP.ang   NP.dist               r        rdot            delta      deldot    S-T-O')
            if i >= istart and i < iend:
                l_s = l.split(',')
                l_s.pop(1)
                l_s.pop(1)
                fb.write(' '.join(l_s))
            else:
                fb.write(l)
                # with open(ephemfile,'w') as fb:
                #     for i,l in enumerate(lines):
                #         fb.write(l.replace('*m','').replace('*t',''))
    return ephemfile


def make_ephem_tb(vis, ephemfile=None, output_casa_table_name=None):
    import os.path

    ephemfromvis = make_ephem(vis)
    if os.path.isfile(ephemfromvis):
        output_dictionary = readJPLephem(ephemfromvis)
        time_range_string = str(int(output_dictionary['earliest']['m0']['value'])) + '-' + str(
            int(output_dictionary['latest']['m0']['value'])) + 'dUTC'
        if not output_casa_table_name:
            output_casa_table_name = '_'.join([output_dictionary['NAME'], time_range_string, 'J2000.tab'])
        if ephemfile is not None:
            tim, ra, dec = read_ephem(ephemfile)
            from scipy import interpolate
            tmjd = output_dictionary['data']['MJD']['value']
            tpmjd = tim.mjd
            fra = interpolate.interp1d(tpmjd, ra, kind='cubic')
            output_dictionary['data']['RA']['data']['value'] = fra(tmjd)
            fdec = interpolate.interp1d(tpmjd, dec, kind='cubic')
            output_dictionary['data']['DEC']['data']['value'] = fdec(tmjd)
        # jplreader.ephem_dict_to_table(output_dictionary, output_casa_table_name)
        ephem_dict_to_table(output_dictionary, output_casa_table_name)
        print('Ephemeris table written to {}'.format(output_casa_table_name))
    else:
        print("sun-ephem-geo.txt not found!")
